from __future__ import annotations

import random
from typing import TYPE_CHECKING

from flow.record import RecordDescriptor
from flow.record.base import Record, parse_def

from dissect.target.helpers.descriptor_extensions import (
    RecordDescriptorExtensionBase,
    TargetRecordDescriptorExtension,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


class ExtendableRecordDescriptor(RecordDescriptor):
    def __init__(self, name: str, fields: Sequence[tuple[str, str]] | None = None):
        """A RecordDescriptor with default fields for dissect targets
        automatically added.
        """

        fields = fields or []
        prepend_default_fields = []
        append_default_fields = []
        extended_fields = []
        self.field_fillers = set()
        self.input_fields = set()
        self.target_fields = fields

        # all classes where we will look for `_default_fields` class
        # property and `_fill_default_fields` method
        classes = list(self.__class__.__bases__)

        for cls in classes:
            if hasattr(cls, "_default_fields"):
                if getattr(cls, "_prepend_fields", False):
                    prepend_default_fields.extend(cls._default_fields)
                else:
                    append_default_fields.extend(cls._default_fields)
            if hasattr(cls, "_fill_default_fields"):
                self.field_fillers.add(cls._fill_default_fields)
            if hasattr(cls, "_input_fields"):
                self.input_fields.update(cls._input_fields)

        if isinstance(fields, RecordDescriptor):
            # Clone fields
            fields = fields.get_field_tuples()
        elif fields is None:
            name, fields = parse_def(name)

        default_field_names = {field_name for _, field_name in prepend_default_fields + append_default_fields}
        extended_fields.extend(prepend_default_fields)

        for field_type, field_name in fields:
            if field_name in default_field_names:
                raise TypeError(f"Default field '{field_name}' is not allowed to be explicitly declared")
            extended_fields.append((field_type, field_name))

        extended_fields.extend(append_default_fields)

        super().__init__(name, fields=extended_fields)

    def __call__(self, *args, **kwargs) -> Record:
        """Generate a record.

        Default fields are prefilled if the _target keyword argument is
        supplied, any explicitly supplied (keyword) arguments for these fields
        are discarded.
        """
        if args:
            raise ValueError("Args are not allowed in ExtendableRecordDescriptor")

        for filler in self.field_fillers:
            kwargs = filler(self, kwargs)

        # cleanup input fields from record keyword arguments
        for input_field in self.input_fields:
            kwargs.pop(input_field, None)

        return super().__call__(*args, **kwargs)


def create_extended_descriptor(
    extensions: Sequence[RecordDescriptorExtensionBase], descriptor_class_name: str | None = None
) -> type[ExtendableRecordDescriptor]:
    class_name = descriptor_class_name or f"CustomExtendedRecordDescriptor{int(random.random() * 100000)}"
    # ExtendableRecordDescriptor must come first, since `ExtendableRecordDescriptor.__init__()` constructor
    # must be executed for extensions to work
    return type(class_name, (ExtendableRecordDescriptor, TargetRecordDescriptorExtension, *extensions), {})


TargetRecordDescriptor = create_extended_descriptor([])


def DynamicDescriptor(types: Sequence[str]) -> RecordDescriptor:
    """Returns a RecordDescriptor with the provided types.

    Plugins that yield records are required to provide their RecordDescriptor
    in the export decorator. However, some plugins dynamically create
    descriptors on the fly. The type of some fields can be known beforehand,
    though. This helper function allows plugins to provide a record
    descriptor that has at least those types, so that they can be used by
    things like an IOC checker, which would look for all plugins that yield
    records with a specific field type.
    """
    if not isinstance(types, (list, tuple)):
        raise TypeError("types must be a list or tuple")

    name = "_".join(types)
    return RecordDescriptor(name, [(t, t) for t in types])


ChildTargetRecord = TargetRecordDescriptor(
    "target/child",
    [
        ("string", "type"),
        ("string", "name"),
        ("path", "path"),
    ],
)


WindowsUserRecord = TargetRecordDescriptor(
    "windows/user",
    [
        ("string", "sid"),
        ("string", "name"),
        ("path", "home"),
    ],
)

COMMON_UNIX_FIELDS = [
    ("string", "name"),
    ("string", "passwd"),
    ("varint", "uid"),
    ("varint", "gid"),
    ("string", "gecos"),
    ("path", "home"),
    ("string", "shell"),
    ("string", "source"),
]

UnixUserRecord = TargetRecordDescriptor(
    "unix/user",
    COMMON_UNIX_FIELDS,
)

MacOSUserRecord = TargetRecordDescriptor(
    "macos/user",
    COMMON_UNIX_FIELDS,
)

IOSUserRecord = TargetRecordDescriptor(
    "ios/user",
    COMMON_UNIX_FIELDS,
)

EmptyRecord = RecordDescriptor(
    "empty",
    [],
)

COMMON_INTERFACE_ELEMENTS = [
    ("string", "name"),
    ("string", "type"),
    ("boolean", "enabled"),
    ("net.ipinterface[]", "cidr"),
    ("net.ipaddress[]", "gateway"),
    ("net.ipaddress[]", "dns"),
    ("string[]", "mac"),
    ("string", "source"),
]


UnixInterfaceRecord = TargetRecordDescriptor(
    "unix/network/interface",
    [
        *COMMON_INTERFACE_ELEMENTS,
        ("boolean", "dhcp_ipv4"),  # NetworkManager allows for dual-stack configurations.
        ("boolean", "dhcp_ipv6"),
        ("datetime", "last_connected"),
        ("varint[]", "vlan"),
        ("string", "configurator"),
    ],
)

WindowsInterfaceRecord = TargetRecordDescriptor(
    "windows/network/interface",
    [
        *COMMON_INTERFACE_ELEMENTS,
        ("varint", "metric"),
        ("stringlist", "search_domain"),
        ("datetime", "first_connected"),
        ("datetime", "last_connected"),
        ("boolean", "dhcp"),
        ("varint", "vlan"),
    ],
)

MacOSInterfaceRecord = TargetRecordDescriptor(
    "macos/network/interface",
    [
        *COMMON_INTERFACE_ELEMENTS,
        ("varint", "interface_service_order"),
        ("boolean", "dhcp"),
        ("varint", "vlan"),
    ],
)


COMMON_APPLICATION_FIELDS = [
    ("datetime", "ts_modified"),
    ("datetime", "ts_installed"),
    ("string", "name"),
    ("string", "version"),
    ("string", "author"),
    ("string", "type"),
    ("path", "path"),
]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import call, patch

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.libvirt import LibvirtLoader
from dissect.target.target import Target
from tests.conftest import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``LibvirtLoader`` when opening a ``Target``."""
    vfs = VirtualFilesystem()
    vfs.map_file("/test.xml", absolute_path("_data/loaders/libvirt/qemu.xml"))
    vfs.map_file_fh("/var/lib/libvirt/images/linux2022.qcow2", None)
    vfs.map_file_fh("/second-disk.qcow2", None)

    path = vfs.path("/test.xml")
    with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
        target = opener(path)
        assert isinstance(target._loader, LibvirtLoader)
        assert target.path == path


def test_loader(tmp_path: Path) -> None:
    """Test that ``LibvirtLoader`` works with local files, absolute and relative."""
    xml_file = tmp_path.joinpath("base")
    xml_file.write_text("not a libvirt file")

    assert LibvirtLoader.detect(xml_file) is False

    xml_file = tmp_path.joinpath("base.xml")
    xml_file.write_text("<domain>")
    assert LibvirtLoader.detect(xml_file) is False

    xml_file = tmp_path.joinpath("base.xml")
    xml_file.write_text("not a libvirt file")

    assert LibvirtLoader.detect(xml_file) is False

    vfs = VirtualFilesystem()
    vfs.map_file("/test.xml", absolute_path("_data/loaders/libvirt/qemu.xml"))
    vfs.map_file_fh("/var/lib/libvirt/images/linux2022.qcow2", None)
    vfs.map_file_fh("/second-disk.qcow2", None)

    loader = loader_open(vfs.path("/test.xml"))
    assert isinstance(loader, LibvirtLoader)

    with patch("dissect.target.container.open") as mock_container_open:
        t = Target()
        loader.map(t)

        assert len(t.disks) == 2
        assert mock_container_open.call_count == 2
        mock_container_open.assert_has_calls(
            [
                call(vfs.path("/var/lib/libvirt/images/linux2022.qcow2")),
                call(vfs.path("/second-disk.qcow2")),
            ]
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.browser.safari import SafariPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_safari_mac(target_macos_users: Target, fs_macos: VirtualFilesystem) -> Target:
    fs_macos.map_dir(
        "/Users/dissect/Library/Safari/",
        absolute_path("_data/plugins/apps/browser/safari/generic/"),
    )
    target_macos_users.add_plugin(SafariPlugin)
    return target_macos_users


def test_safari_history(target_safari_mac: Target) -> None:
    records = list(target_safari_mac.safari.history())

    assert len(records) == 5
    assert {"safari"} == {r.browser for r in records}

    # First visit: direct navigation to example.com
    r = records[0]
    assert r.ts == dt("2024-01-01T00:00:00+00:00")
    assert r.url == "https://example.com/"
    assert r.title == "Example Domain"
    assert r.host == "example"
    assert r.visit_count == 3
    assert not r.hidden
    assert r.from_visit is None
    assert r.from_url is None

    # Second visit: redirect from visit 1
    r = records[1]
    assert r.ts == dt("2024-01-01T00:01:00+00:00")
    assert r.url == "https://example.com/page"
    assert r.title == "Example Page"
    assert r.from_visit == 1
    assert r.from_url == "https://example.com/"

    # Fifth visit: synthesized (hidden)
    r = records[4]
    assert r.hidden
    assert r.title is None



from __future__ import annotations

import urllib.parse
import warnings
from typing import TYPE_CHECKING, Final

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import Loader
from dissect.target.plugin import arg

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


@arg("--log-hint", dest="hint", help="hint for file type")
class LogLoader(Loader):
    """Load separate log files without a target.

    Usage:

    ``target-query /evtx/* -L log -f evtx``

    """

    LOGS_DIRS: Final[dict[str, str]] = {
        "evtx": "sysvol/windows/system32/winevt/logs",
        "evt": "sysvol/windows/system32/config",
        "iis": "sysvol/files/logs/",
    }

    def __init__(self, path: Path, parsed_path: urllib.parse.ParseResult | None = None):
        super().__init__(path, parsed_path=parsed_path)
        warnings.warn(
            "The LogLoader is deprecated in favor of direct files (`--direct`)"
            " and will be removed in dissect.target 3.24",
            FutureWarning,
            stacklevel=2,
        )
        self.options = dict(urllib.parse.parse_qsl(parsed_path.query, keep_blank_values=True)) if parsed_path else {}

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    def map(self, target: Target) -> None:
        vfs = VirtualFilesystem(case_sensitive=False, alt_separator=target.fs.alt_separator)
        target.filesystems.add(vfs)
        target.fs.mount("/", vfs)
        for entry in self.path.parent.glob(self.path.name):
            ext = self.options.get("hint", entry.suffix.lower()).strip(".")
            if (mapping := self.LOGS_DIRS.get(ext, None)) is None:
                continue
            mapping = str(vfs.path(mapping).joinpath(entry.name))
            vfs.map_file(mapping, str(entry))



from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import RegistryKeyNotFoundError
from dissect.target.helpers.regutil import VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.log import evt
from dissect.target.plugins.scrape import scrape
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import VirtualFilesystem


def mock_registry_log_location(target_win: Target, reg_key_name: str, mock_log_path: str) -> None:
    # Mock eventlog registry key in a specific control set,
    # CurrentControlSet used in preconfigured value is a soft link
    registry_hive = target_win.registry._root

    eventlog_key = evt.WindowsEventlogsMixin.EVENTLOG_REGISTRY_KEY.replace("CurrentControlSet", "ControlSet001")

    try:
        registry_hive.key(eventlog_key)
    except RegistryKeyNotFoundError:
        registry_hive.map_key(eventlog_key, VirtualKey(registry_hive, eventlog_key))

    # Set 'File' registry value to eventlog file path
    registry_key = f"{eventlog_key}\\{reg_key_name}"
    registry_subkey = VirtualKey(registry_hive, registry_key)
    registry_hive.map_key(registry_key, registry_subkey)

    eventlog_security_value = VirtualValue(registry_hive, "File", mock_log_path)
    registry_hive.map_value(registry_key, "File", eventlog_security_value)


@pytest.mark.parametrize(
    ("is_in_directory", "is_in_registry"),
    [
        (True, False),
        (False, True),
        (True, True),
    ],
)
def test_evt_plugin(
    target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path, is_in_directory: bool, is_in_registry: bool
) -> None:
    target_win.add_plugin(evt.EvtPlugin)

    evt_log_file = absolute_path("_data/plugins/os/windows/log/evt/TestLog.evt")
    expected_records = 0

    if is_in_directory:
        evt_dir_file = tmp_path / "TestLogDir.evt"
        shutil.copyfile(evt_log_file, evt_dir_file)

        fs_win.map_file("windows/system32/config/TestLog.evt", evt_dir_file)
        expected_records += 5

    if is_in_registry:
        evt_reg_file = tmp_path / "TestLogReg.evt"
        shutil.copyfile(evt_log_file, evt_reg_file)

        # Set a log path in the registry key and map that path
        mock_log_path = "somefolder/TestLogMock.EVT"
        mock_registry_log_location(target_win, "Security", f"C:\\{mock_log_path}")
        fs_win.map_file(mock_log_path, evt_reg_file)

        # Set path that should be skipped
        mock_registry_log_location(target_win, "MockApp", "C:\\somefolder\\BadFile.wrong")

        # Set path that doesn't exist
        mock_registry_log_location(target_win, "MockApp", "C:\\somefolder\\BadFile.evt")

        expected_records += 5

    records = list(target_win.evt())

    assert len(records) == expected_records

    # Verify that out of all received records, only 5 are unique
    assert len({str(rec.ts) for rec in records}) == 5


def test_evt_scraping(target_win: Target) -> None:
    target_win.add_plugin(scrape.ScrapePlugin)

    plugin = evt.EvtPlugin(target_win)
    evt_log_file = absolute_path("_data/plugins/os/windows/log/evt/TestLog.evt")

    with evt_log_file.open("rb") as fh:
        target_win.disks.add(fh)
        # Make sure that non-zero initial position does not break scraping
        fh.seek(200)
        scraped_records = list(plugin.scraped_evt())

    assert len(scraped_records) == 5


def test_evt_direct_mode() -> None:
    data_path = absolute_path("_data/plugins/os/windows/log/evt/TestLog.evt")

    target = Target.open_direct([data_path])
    records = list(target.evt())

    assert len(records) == 5



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers import configutil
from dissect.target.helpers.record import EmptyRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class AndroidPlugin(LinuxPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self.target = target

        self.build_prop_paths = set(find_build_props(self.target.fs))
        self.props = {}

        for build_prop in self.build_prop_paths:
            try:
                self.props.update(
                    configutil.parse(
                        build_prop,
                        hint="meta_bare",
                        separator=("=",),
                        comment_prefixes=("#",),
                    ).parsed_data
                )
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Unable to parse Android build.prop file %s: %s", build_prop, e)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        ANDROID_PATHS = (
            "data",
            "system",
            "vendor",
            "product",
        )

        for fs in target.filesystems:
            if all(fs.exists(p) for p in ANDROID_PATHS) and any(find_build_props(fs)):
                return fs
        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> str | None:
        return self.props.get("ro.build.host")

    @export(property=True)
    def ips(self) -> list[str]:
        return []

    @export(property=True)
    def version(self) -> str:
        full_version = "Android"

        release_version = self.props.get("ro.build.version.release")
        if release_version := self.props.get("ro.build.version.release"):
            full_version += f" {release_version}"

        if security_patch_version := self.props.get("ro.build.version.security_patch"):
            full_version += f" ({security_patch_version})"

        return full_version

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.ANDROID.value

    @export(record=EmptyRecord)
    def users(self) -> Iterator[EmptyRecord]:
        yield from ()


def find_build_props(fs: Filesystem) -> Iterator[Path]:
    """Search for Android ``build.prop`` files on the provided :class:`Filesystem`."""
    if (root_prop := fs.path("/build.prop")).is_file():
        yield root_prop

    for prop in fs.path("/").glob("*/build.prop"):
        if prop.is_file():
            yield prop



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.loader import Loader
from dissect.target.loaders.dir import map_dirs
from dissect.target.loaders.tar import TarSubLoader
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    import tarfile as tf
    from pathlib import Path

    from dissect.target.target import Target


# From tech support readme
# Files/Directories of Interest:
# ------------------------------
#
# error.log: log containing errors that ocurred while running vm-support
# action.log: a log of all commands, and/or actions run during vm-support
# commands: a directory containing output files for all commands run
#
# All other directories and files should be mirrors of the the ESXi system
# vm-support was run on.

EXPECTED_FILES_OR_DIR = ["etc/vmware/esx.conf", "error.log", "action.log"]


class VmSupportLoader(Loader):
    """Loader for extracted ESXi vm-support.

    References:
        - https://knowledge.broadcom.com/external/article/313542
    """

    @staticmethod
    def detect(path: Path) -> bool:
        if not path.is_dir():
            return False
        root_dir = next(path.iterdir())
        return all(root_dir.joinpath(f).exists() for f in EXPECTED_FILES_OR_DIR)

    def map(self, target: Target) -> None:
        map_dirs(target, [next(self.absolute_path.iterdir())], OperatingSystem.ESXI)


class VmSupportTarSubloader(TarSubLoader):
    """Loader for tar-based ESXi vm-support.

    References:
        - https://knowledge.broadcom.com/external/article/313542
    """

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        if not (names := tarfile.getnames()):
            return False
        root = names[0].split("/")[0]
        required_paths = {f"{root}/{f}" for f in EXPECTED_FILES_OR_DIR}
        return required_paths.issubset(names)

    def map(self, target: Target) -> None:
        fs = TarFilesystem(self.tar.fileobj, tarfile=self.tar, base=self.tar.getnames()[0].split("/")[0])
        target.filesystems.add(fs)



#!/usr/bin/env python
from __future__ import annotations

import argparse

from dissect.target.exceptions import TargetError
from dissect.target.helpers.logging import get_logger
from dissect.target.plugins.filesystem.yara import HAS_YARA, YaraPlugin
from dissect.target.tools.query import record_output
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    open_targets,
    process_generic_arguments,
)

log = get_logger(__name__)


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-yara",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )

    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to load")
    parser.add_argument("-s", "--strings", action="store_true", help="print output as string")

    plugin_parser = parser.add_argument_group("YARA plugin options")
    for args, kwargs in getattr(YaraPlugin.yara, "__args__", []):
        plugin_parser.add_argument(*args, **kwargs)

    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(parser, args)

    if not HAS_YARA:
        log.error("yara-python is not installed: pip install yara-python")
        return 1

    if not args.targets:
        log.error("No targets provided")
        return 1

    try:
        for target in open_targets(args):
            rs = record_output(args.strings, False)
            for record in target.yara(args.rules, args.path, args.max_size, args.check, args.no_decompress):
                rs.write(record)

    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

import io
from typing import TYPE_CHECKING
from unittest.mock import Mock, call, patch

from dissect.target.containers.raw import RawContainer
from dissect.target.plugins.scrape.scrape import ScrapePlugin
from dissect.target.volume import EncryptedVolumeSystem, LogicalVolumeSystem

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_create_streams(target_bare: Target) -> None:
    target_bare.add_plugin(ScrapePlugin)

    streams = list(target_bare.scrape.create_streams())
    assert len(streams) == 0

    mock_disk = Mock(name="mock_disk")
    mock_disk.size = 1024 * 64

    mock_volume_1 = Mock(name="mock_volume_1")
    mock_volume_1.disk = mock_disk
    mock_volume_1.offset = 1024
    mock_volume_1.size = 1024 * 8
    mock_volume_2 = Mock(name="mock_volume_2")
    mock_volume_2.disk = mock_disk
    mock_volume_2.offset = 1024 * 10
    mock_volume_2.size = 1024 * 8

    mock_disk.vs.volumes = [mock_volume_1, mock_volume_2]

    target_bare.disks.add(mock_disk)

    expected_base = [
        call(0, 1024, mock_disk, 0),
        call(1024, 1024 * 8, mock_volume_1, 0),
        call(1024 * 9, 1024, mock_disk, 1024 * 9),
        call(1024 * 10, 1024 * 8, mock_volume_2, 0),
        call(1024 * 18, 1024 * 46, mock_disk, 1024 * 18),
    ]

    with patch("dissect.util.stream.MappingStream.add") as mock_add:
        streams = list(target_bare.scrape.create_streams())
        mock_add.assert_has_calls(expected_base)

    mock_encrypted_volume = Mock(name="mock_encrypted_volume")
    mock_encrypted_volume.disk = mock_volume_1
    mock_encrypted_volume.offset = 1024
    mock_encrypted_volume.size = 1024 * 8
    mock_encrypted_volume.vs = Mock(spec=EncryptedVolumeSystem)

    target_bare.volumes.entries = [mock_encrypted_volume]

    with patch("dissect.util.stream.MappingStream.add") as mock_add:
        streams = list(target_bare.scrape.create_streams(encrypted=False))
        mock_add.assert_has_calls(expected_base)

        mock_add.reset_mock()

        streams = list(target_bare.scrape.create_streams(encrypted=True))
        mock_add.assert_has_calls(
            [
                call(0, 1024, mock_disk, 0),
                call(1024, 1024 * 8, mock_encrypted_volume, 0),
                call(1024 * 9, 1024, mock_disk, 1024 * 9),
                call(1024 * 10, 1024 * 8, mock_volume_2, 0),
                call(1024 * 18, 1024 * 46, mock_disk, 1024 * 18),
            ]
        )

        mock_add.reset_mock()

        streams = list(target_bare.scrape.create_streams(encrypted=True, all=True))
        mock_add.assert_has_calls(
            [
                call(0, 1024, mock_disk, 0),
                call(1024, 1024 * 8, mock_volume_1, 0),
                call(1024 * 9, 1024, mock_disk, 1024 * 9),
                call(1024 * 10, 1024 * 8, mock_volume_2, 0),
                call(1024 * 18, 1024 * 46, mock_disk, 1024 * 18),
                call(0, 1024 * 8, mock_encrypted_volume, 0),
            ]
        )

    mock_lvm_volume = Mock(name="mock_lvm_volume")
    mock_lvm_volume.size = 1024 * 16
    mock_lvm_volume.disk = [mock_volume_1, mock_volume_2]
    mock_lvm_volume.vs = Mock(spec=LogicalVolumeSystem)

    target_bare.volumes.entries = [mock_lvm_volume]

    with patch("dissect.util.stream.MappingStream.add") as mock_add:
        streams = list(target_bare.scrape.create_streams(lvm=False))
        mock_add.assert_has_calls(expected_base)

        mock_add.reset_mock()

        streams = list(target_bare.scrape.create_streams(lvm=True))
        mock_add.assert_has_calls(
            [
                call(0, 1024, mock_disk, 0),
                call(1024 * 9, 1024, mock_disk, 1024 * 9),
                call(1024 * 18, 1024 * 46, mock_disk, 1024 * 18),
                call(0, 1024 * 16, mock_lvm_volume, 0),
            ]
        )

        mock_add.reset_mock()

        streams = list(target_bare.scrape.create_streams(lvm=True, all=True))
        mock_add.assert_has_calls(
            [
                call(0, 1024, mock_disk, 0),
                call(1024, 1024 * 8, mock_volume_1, 0),
                call(1024 * 9, 1024, mock_disk, 1024 * 9),
                call(1024 * 10, 1024 * 8, mock_volume_2, 0),
                call(1024 * 18, 1024 * 46, mock_disk, 1024 * 18),
                call(0, 1024 * 16, mock_lvm_volume, 0),
            ]
        )


def test_find(target_bare: Target) -> None:
    target_bare.add_plugin(ScrapePlugin)

    buf = (b"\x00" * 1024 * 64) + (b"ABCD" + b"\x00" * ((1024 * 4) - 4)) + (b"\x00" * 1024 * 60)
    mock_disk = RawContainer(io.BytesIO(buf))
    target_bare.disks.add(mock_disk)

    mock_progress = Mock()
    for disk, stream, needle, offset, match in target_bare.scrape.find(b"ABCD", progress=mock_progress):
        assert disk.size == 1024 * 128
        assert stream.size == 1024 * 128
        assert needle == b"ABCD"
        assert offset == 1024 * 64
        assert not match  # only for regex matches

    for i in range(0, 1024 * 128, 8192):
        mock_progress.assert_any_call(mock_disk, i, 1024 * 128)



from __future__ import annotations

import io
import re
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence import EWF
from dissect.evidence.ewf import find_files

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class EwfContainer(Container):
    """Expert Witness Disk Image Format."""

    __type__ = "ewf"

    def __init__(self, fh: list | BinaryIO | Path, *args, **kwargs):
        fhs = [fh] if not isinstance(fh, list) else fh
        if hasattr(fhs[0], "read"):
            self.ewf = EWF(fhs)
        else:
            self.ewf = EWF(find_files(fhs[0]))

        self._stream = self.ewf.open()
        super().__init__(fh, self.ewf.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(3) in (b"EVF", b"LVF", b"LEF")

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return re.match(r"\.[EeLs]x?01$", path.suffix)

    def read(self, length: int) -> bytes:
        return self._stream.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self._stream.seek(offset, whence)

    def tell(self) -> int:
        return self._stream.tell()

    def close(self) -> None:
        self._stream.close()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.credential.credential import WindowsCredentialPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

WinlogonRecord = TargetRecordDescriptor(
    "windows/credential/winlogon",
    [
        ("datetime", "ts_mtime"),
        ("string", "password"),
        ("path", "source"),
    ],
)


class WinlogonPlugin(WindowsCredentialPlugin):
    """Windows Winlogon plugin."""

    __namespace__ = "winlogon"

    WINLOGON_KEY = "HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Windows registry plugin not available on target")

    @export(record=WinlogonRecord)
    def winlogon(self) -> Iterator[WinlogonRecord]:
        """Yield Windows Winlogon DefaultPassword strings.

        Extracts plaintext ``DefaultPassword`` values from the ``Winlogon`` registry.

        References:
            - https://learn.microsoft.com/en-us/troubleshoot/windows-server/user-profiles-and-logon/turn-on-automatic-logon
        """

        for key in self.target.registry.keys(self.WINLOGON_KEY):
            try:
                yield WinlogonRecord(
                    ts_mtime=key.ts,
                    password=key.value("DefaultPassword").value,
                    source=self.WINLOGON_KEY,
                    _target=self.target,
                )
            except RegistryValueNotFoundError:  # noqa: PERF203
                pass



from __future__ import annotations

from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, NamedTuple, Union

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record import (
    IOSUserRecord,
    MacOSUserRecord,
    UnixUserRecord,
    WindowsUserRecord,
)
from dissect.target.plugin import InternalPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

UserRecord = Union[UnixUserRecord, WindowsUserRecord, MacOSUserRecord, IOSUserRecord]  # noqa: UP007


class UserDetails(NamedTuple):
    user: UserRecord
    home_path: TargetPath | None


class UsersPlugin(InternalPlugin):
    """Internal plugin that provides helper functions for retrieving user details."""

    __namespace__ = "user_details"

    def __init__(self, target: Target):
        super().__init__(target)
        self.find = lru_cache(32)(self.find)

    def check_compatible(self) -> None:
        if not hasattr(self.target, "users"):
            raise UnsupportedPluginError("Unsupported Plugin")

    def find(
        self,
        sid: str | None = None,
        uid: int | None = None,
        username: str | None = None,
        force_case_sensitive: bool = False,
    ) -> UserDetails | None:
        """Find user record by ``sid``, ``uid`` or ``username`` and return :class:`UserDetails` object."""
        if all(x is None for x in [sid, uid, username]):
            raise ValueError("Either sid or uid or username is expected")

        def is_name_matching(name: str) -> bool:
            if force_case_sensitive or self.target.os != "windows":
                # always do case-sensitive match for non-Windows OSes
                return name == username
            return name.lower() == username.lower()

        for user in self.target.users():
            if (
                (sid is not None and user.sid == sid)
                or (uid is not None and user.uid == uid)
                or (username is not None and is_name_matching(user.name))
            ):
                return self.get(user)

        return None

    def get(self, user: UserRecord) -> UserDetails:
        """Return additional details about the user."""
        # Resolving the user home can not use the user's environment variables,
        # as those depend on the user's home to be known first. So we resolve
        # without providing the user (s)id.
        home_path = self.target.resolve(str(user.home)) if user.home else None
        return UserDetails(user=user, home_path=home_path)

    def all(self) -> Iterator[UserDetails]:
        """Return :class:`UserDetails` objects for all users found."""
        for user in self.target.users():
            yield self.get(user)

    def all_with_home(self) -> Iterator[UserDetails]:
        """Return :class:`UserDetails` objects for users that have existing directory set as home directory."""

        seen: set[TargetPath] = set()
        try:
            for user in self.target.users():
                if user.home:
                    user_details = self.get(user)
                    if user_details.home_path.exists():
                        seen.add(user_details.home_path)
                        yield user_details
        except Exception as e:
            self.target.log.warning("Failed to retrieve user details, falling back to hardcoded locations")
            self.target.log.debug("", exc_info=e)

        # Iterate over misc home directories
        for misc_home_dir, user_criterion in self.target.misc_user_paths():
            # We skip the entry if no user can be found.
            # We cannot use set membership check here because of https://github.com/fox-it/dissect.target/issues/1203
            if any(seen_path.samefile(misc_home_dir) for seen_path in seen) or not user_criterion:
                continue

            if (user_details := self.find(**{user_criterion[0]: user_criterion[1]})) is None:
                continue

            yield UserDetails(user=user_details.user, home_path=misc_home_dir)
            seen.add(misc_home_dir)

    @cached_property
    def all_home_paths(self) -> Iterator[TargetPath]:
        """Return all home directories of users, including miscellaneous user directories that may not be linked to discovered local users."""  # noqa: E501

        seen: set[TargetPath] = set()
        try:
            for user in self.target.users():
                if user.home and (home_path := self.target.resolve(str(user.home))).exists():
                    yield home_path
                    seen.add(home_path)
        except Exception as e:
            self.target.log.warning("Failed to retrieve user details, falling back to hardcoded locations")
            self.target.log.debug("", exc_info=e)

        # Iterate over misc home directories
        for misc_home_dir, _ in self.target.misc_user_paths():
            # We cannot use set membership check here because of https://github.com/fox-it/dissect.target/issues/1203
            if not any(seen_path.samefile(misc_home_dir) for seen_path in seen):
                yield misc_home_dir
                seen.add(misc_home_dir)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

TrendMicroWFLogRecord = TargetRecordDescriptor(
    "application/av/trendmicro/wf/log",
    [
        ("datetime", "ts"),
        ("string", "threat"),
        ("path", "path"),
        ("varint", "lineno"),
    ],
)


TrendMicroWFFirewallRecord = TargetRecordDescriptor(
    "application/av/trendmicro/wf/firewall",
    [
        ("datetime", "ts"),
        ("net.ipaddress", "local_ip"),
        ("net.ipaddress", "remote_ip"),
        ("string", "direction"),
        ("uint16", "port"),
        ("path", "path"),
        ("string", "description"),
    ],
)


pfwlog_def = """
struct firewall_entry {
    char      _pad1[1];
    char      direction;
    uint16    port;
    uint32    timestamp;
    char      _pad2[8];
    char      local_ip[65];
    char      remote_ip[65];
    char      path[520];
    wchar     description[128];
    char      _pad3[10];
};
"""
c_pfwlog = cstruct().load(pfwlog_def)


class TrendMicroPlugin(Plugin):
    """TrendMicro antivirus plugin."""

    __namespace__ = "trendmicro"

    LOG_FOLDER = "sysvol/Program Files (x86)/Trend Micro/Security Agent"
    LOG_FILE_FIREWALL = f"{LOG_FOLDER}/PFW/PfwLog_*.dat"  # Windows intrusions
    LOG_FILE_INFECTIONS = f"{LOG_FOLDER}/Misc/pccnt35.log"  # Windows infections

    def __init__(self, target: Target):
        super().__init__(target)
        self.codepage = self.target.codepage or "ascii"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.LOG_FOLDER).exists():
            raise UnsupportedPluginError

    @export(record=TrendMicroWFLogRecord)
    def wflogs(self) -> Iterator[TrendMicroWFLogRecord]:
        """Return Trend Micro Worry-free log history records.

        Yields TrendMicroWFLogRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): timestamp.
            threat (string): Description of the detected threat.
            path (string): Path to file that is associated with the threat.
            filename (string): Name to file that is associated with the threat.
            lineno (uint16): Line number for reference for further investigation.
        """
        with self.target.fs.path(self.LOG_FILE_INFECTIONS).open("rt", 0, self.codepage) as f:
            for lineno, line in enumerate(f.readlines()):
                cells = line.split("<;>")
                yield TrendMicroWFLogRecord(
                    ts=from_unix(int(cells[9])),
                    threat=cells[2],
                    path=self.target.fs.path(cells[6] + cells[7]),
                    lineno=lineno,
                )

    @export(record=TrendMicroWFFirewallRecord)
    def wffirewall(self) -> Iterator[TrendMicroWFFirewallRecord]:
        """Return Trend Micro Worry-free firewall log history records.

        Yields TrendMicroWFFirewallRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): timestamp.
            local_ip (net.ipadress): Local IPv4/IPv6.
            remote_ip (net.ipaddress): Remote IPv4/IPv6.
            port (uint16): Port of suspicious connection.
            direction (string): Direction of the traffic
            path (string): Path to object that initiated/received connection
            description (string): Description of the detected threat
        """
        for firewall_log in self.target.fs.glob_ext(self.LOG_FILE_FIREWALL):
            with firewall_log.open() as log:
                try:
                    while entry := c_pfwlog.firewall_entry(log):
                        yield TrendMicroWFFirewallRecord(
                            ts=from_unix(entry.timestamp),
                            local_ip=entry.local_ip.strip(b"\x00").decode(self.codepage),
                            remote_ip=entry.remote_ip.strip(b"\x00").decode(self.codepage),
                            port=entry.port,
                            direction=("out" if entry.direction == b"\x01" else "in"),
                            path=self.target.fs.path(entry.path.strip(b"\x00").decode(self.codepage)),
                            description=entry.description.strip("\x00"),
                        )
                except EOFError:
                    pass



from __future__ import annotations

import codecs
import re
from typing import TYPE_CHECKING

from defusedxml import ElementTree
from dissect.util.ts import wintimestamp
from flow.record.base import is_valid_field_name

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

CAMEL_CASE_PATTERNS = [re.compile(r"(\S)([A-Z][a-z]+)"), re.compile(r"([a-z0-9])([A-Z])"), re.compile(r"(\w)[.\s](\w)")]
RE_VALID_KEY_START_CHARS = re.compile(r"[a-zA-Z]")
RE_VALID_KEY_CHARS = re.compile(r"[a-zA-Z0-9_]")


def _create_record_descriptor(record_name: str, record_fields: list[tuple[str, str]]) -> TargetRecordDescriptor:
    record_fields.extend(
        [
            ("path", "wer_file_path"),
            ("path", "metadata_file_path"),
        ]
    )
    return TargetRecordDescriptor(record_name, record_fields)


class WindowsErrorReportingPlugin(Plugin):
    """Plugin for parsing Windows Error Reporting files."""

    WER_LOG_DIRS = (
        "sysvol/ProgramData/Microsoft/Windows/WER/ReportArchive",
        "sysvol/ProgramData/Microsoft/Windows/WER/ReportQueue",
        "%userprofile%/AppData/Local/Microsoft/Windows/WER/ReportArchive",
        "%userprofile%/AppData/Local/Microsoft/Windows/WER/ReportQueue",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.wer_files = [
            res.iterdir()
            for wer_dir in self.WER_LOG_DIRS
            for res in self.target.fs.path(wer_dir).rglob("*")
            if res.is_dir()
        ]

    def check_compatible(self) -> None:
        if not self.wer_files:
            raise UnsupportedPluginError("No Windows Error Reporting directories found.")

    def _sanitize_key(self, key: str) -> str:
        # Convert camel case to snake case
        for pattern in CAMEL_CASE_PATTERNS:
            key = pattern.sub(r"\1_\2", key)

        clean_key = ""
        separator = "_"
        prev_encoded = False
        for idx, char in enumerate(key):
            if prev_encoded:
                clean_key += separator
            clean_key += char
            if not is_valid_field_name(clean_key):
                clean_key = clean_key[:-1]
                prefix = f"{separator}x" if (idx != 0 and not prev_encoded) else "x"
                clean_key += prefix + char.encode("utf-8").hex()
                prev_encoded = True
            else:
                prev_encoded = False

        return clean_key.lower()

    def _collect_wer_data(self, wer_file: Path) -> tuple[list[tuple[str, str]], dict[str, str]]:
        """Parse data from a .wer file."""
        record_values = {}
        record_fields = []
        key = None

        # Default encoding when no BOM is present
        encoding = "utf-16-le"

        # If a BOM header is present we can decode it using utf-16
        with wer_file.open("rb") as fh:
            if fh.read(len(codecs.BOM)) in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
                encoding = "utf-16"

        for line in wer_file.read_text(encoding).splitlines():
            if len(line_split := line.rstrip().split("=", 1)) != 2:
                continue

            name, value = line_split
            record_type = "string"

            # Dynamic entry with key and value on separate lines
            if "].Name" in name and not key:
                key = value
                # Set key and continue to get value on the next line
                continue

            # Dynamic entry with key and value on the same line
            if "]." in name and not key:
                category, name = name.split(".", 1)
                key = f"{category.split('[')[0]}{name}"

            if "EventTime" in name:
                value = wintimestamp(int(value))
                record_type = "datetime"
                key = "ts"

            key = self._sanitize_key(key if key else name)
            if not key:
                self.target.log.warning("Sanitizing key resulted in empty key, skipping line %r", line)
                key = None
                continue

            if key in record_values:
                self.target.log.warning("Key does already exist, skipping line %r", line)
                key = None
                continue

            record_values[key] = value
            record_fields.append((record_type, key)) if key != "ts" else record_fields.insert(0, (record_type, key))
            # Reset key necessary for dynamic entries and ts
            key = None

        return record_fields, record_values

    def _collect_wer_metadata(self, metadata_xml_file: Path) -> tuple[list[tuple[str, str]], dict[str, str]]:
        """Parse data from a metadata .xml file linked to a .wer file."""
        record_fields = []
        record_values = {}
        file = metadata_xml_file.read_text("utf-16")

        tree = ElementTree.fromstring(file)
        for metadata in tree.iter("WERReportMetadata"):
            for category in metadata:
                for value in category:
                    if not (record_value := value.text.strip("\t\n")):
                        continue

                    key = self._sanitize_key(f"{category.tag}{value.tag}")
                    if not key:
                        self.target.log.warning("Sanitizing key resulted in empty key, skipping value %r", value)
                        continue

                    if key in record_values:
                        self.target.log.warning("Key already exists, skipping value %r", value)
                        continue

                    record_fields.append(("string", key))
                    record_values[key] = record_value

        return record_fields, record_values

    @export(record=DynamicDescriptor(["path", "string", "datetime"]))
    def wer(self) -> Iterator[DynamicDescriptor]:
        """Return information from Windows Error Reporting (WER) files.

        Windows Error Reporting (WER) is used by Microsoft to create a report when an application crashes. These reports
        can be sent to Microsoft, on which basis Microsoft can provide the user with troubleshooting information. Since
        malware usually crashes more often than legitimate software, the presence of these WER files and/or the
        information within may be useful for analysis. For example, it may contain the file hash of the crashed
        application within the target_app_id field.

        References:
            - https://learn.microsoft.com/en-us/windows/win32/wer/windows-error-reporting
            - https://medium.com/dfir-dudes/amcache-is-not-alone-using-wer-files-to-hunt-evil-86bdfdb216d7

        Yields dynamically created records based on the fields in the files. A record at least contains the following
        fields:

        .. code-block:: text

            ts (datetime): The moment in time when the error event took place.
            version (string): WER file version.
            event_type (string): WER file event type.
            consent (string): WER file consent to be sent to Microsoft.
            report_identifier (string): WER file report identifier
            app_session_guid (string): GUID for the app session causing/reporting the error.
            target_app_id (string): WER file target app ID which may contain the application hash.
            target_app_ver (string): WER file target app version.
            boot_id (string): WER file boot ID.
            response_type (string): WER file response type.
            friendly_event_name (string): Human readable event name.
            app_name (string): WER file application name.
            app_path (string): Path to application that caused/reported the error.
            report_description (string): WER file report description.
            application_identity (string): WER file application identity.
            metadata_hash (string): WER file metadata hash.
            wer_file_path (path): File path to the WER report on the target.
            metadata_file_path (path): File path to the metadata XML file on the target (if present).
        """
        for files in self.wer_files:
            record_fields = []
            record_values = {"_target": self.target}

            for file in files:
                if file.suffix == ".wer":
                    record_values["wer_file_path"] = file
                    wer_report_fields, wer_report_values = self._collect_wer_data(file)
                    # make sure wer_report_fields are the first entries in the list
                    record_fields = wer_report_fields + record_fields
                    record_values = record_values | wer_report_values
                elif ".WERInternalMetadata" in file.suffixes:
                    record_values["metadata_file_path"] = file
                    metadata_fields, metadata_values = self._collect_wer_metadata(file)
                    record_fields.extend(metadata_fields)
                    record_values = metadata_values | record_values

            record = _create_record_descriptor("filesystem/windows/wer/report", record_fields)
            yield record(**record_values)



from __future__ import annotations

import operator
import uuid
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.remoteaccess.anydesk import AnydeskPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_anydesk_plugin_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "ProgramData/AnyDesk/TestAnydesk.trace",
        absolute_path("_data/plugins/apps/remoteaccess/anydesk/TestAnydesk.trace"),
    )

    target_win_users.add_plugin(AnydeskPlugin)

    records = list(target_win_users.anydesk.logs())
    assert len(records) == 1

    assert records[0].ts == datetime(2021, 11, 11, 12, 34, 56, 789000, tzinfo=timezone.utc)
    assert records[0].message == "LEVEL Strip the headers, trace the source!"
    assert records[0].source == "sysvol/ProgramData/AnyDesk/TestAnydesk.trace"
    assert records[0].username is None
    assert records[0].user_id is None
    assert records[0].user_home is None


def test_anydesk_plugin_user_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "Users/John/AppData/Roaming/AnyDesk/TestAnydesk.trace",
        absolute_path("_data/plugins/apps/remoteaccess/anydesk/TestAnydesk.trace"),
    )

    target_win_users.add_plugin(AnydeskPlugin)

    records = list(target_win_users.anydesk.logs())
    records.sort(key=operator.attrgetter("source"))
    assert len(records) == 1

    user_details = target_win_users.user_details.find(username="John")
    assert records[0].ts == datetime(2021, 11, 11, 12, 34, 56, 789000, tzinfo=timezone.utc)
    assert records[0].message == "LEVEL Strip the headers, trace the source!"
    assert records[0].source == "C:/Users/John/AppData/Roaming/AnyDesk/TestAnydesk.trace"
    assert records[0].username == user_details.user.name
    assert records[0].user_id == user_details.user.sid
    assert records[0].user_home == user_details.user.home


def test_anydesk_plugin_multiple_trace_files(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    trace_file = absolute_path("_data/plugins/apps/remoteaccess/anydesk/Another.trace")
    trace_files = 0

    for str_path in AnydeskPlugin.SERVICE_GLOBS:
        if "var/log/" in str_path:
            continue
        fs_win.map_file(str_path.replace("sysvol/", "").replace("*", uuid.uuid4().hex[:6]), trace_file)
        trace_files += 1

    for str_path in AnydeskPlugin.USER_GLOBS:
        if str_path.startswith(".anydesk"):
            continue
        fs_win.map_file("Users\\John\\" + str_path.replace("*", uuid.uuid4().hex[:6]), trace_file)
        trace_files += 1

    target_win_users.add_plugin(AnydeskPlugin)
    records = list(target_win_users.anydesk.logs())

    assert len(target_win_users.anydesk.trace_files) == trace_files
    assert len(records) == trace_files * 419


def test_anydesk_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    trace_file = absolute_path("_data/plugins/apps/remoteaccess/anydesk/Another.trace")
    fs_unix.map_file("/var/log/anydesk/1337.trace", trace_file)
    fs_unix.map_file("/root/.anydesk/1337.trace", trace_file)
    fs_unix.map_file("/root/.anydesk_ad_1337/1337.trace", trace_file)

    target_unix_users.add_plugin(AnydeskPlugin)
    records = list(target_unix_users.anydesk.logs())

    assert len(target_unix_users.anydesk.trace_files) == 3
    assert len(records) == 3 * 419


def test_anydesk_plugin_filetransfer(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "ProgramData/AnyDesk/file_transfer_trace.txt",
        absolute_path("_data/plugins/apps/remoteaccess/anydesk/file_transfer_trace.txt"),
    )

    target_win_users.add_plugin(AnydeskPlugin)

    records = list(target_win_users.anydesk.filetransfer())
    assert len(records) == 2

    assert records[0].ts == datetime(2025, 1, 1, 13, 37, tzinfo=timezone.utc)
    assert records[0].message == "Clipboard  2025-01-01, 13:37 start  download 'malware.exe' (~0 B out of 4.20 MiB)"
    assert records[0].source == "sysvol/ProgramData/AnyDesk/file_transfer_trace.txt"
    assert records[0].filename == "malware.exe"
    assert records[0].username is None
    assert records[0].user_id is None
    assert records[0].user_home is None



from __future__ import annotations

from types import ModuleType
from unittest.mock import patch

import pytest

from dissect.target.helpers.lazy import import_lazy


def test_working() -> None:
    mock_module = ModuleType("mock")
    mock_module.function = lambda: "foo"

    with patch("importlib.import_module") as mock_import_module:
        mock_import_module.return_value = mock_module

        lazy_mod = import_lazy("mock")
        lazy_attr = lazy_mod.function

        assert repr(lazy_mod) == "<lazy mock loaded=False failed=False>"
        assert repr(lazy_attr) == "<lazyattr mock.function loaded=False failed=False>"

        assert lazy_attr() == "foo"

        assert repr(lazy_mod) == "<lazy mock loaded=True failed=False>"
        assert repr(lazy_attr) == "<lazyattr mock.function loaded=True failed=False>"


def test_failing_import() -> None:
    with patch("importlib.import_module") as mock_import_module:
        mock_import_module.side_effect = ImportError("Module not found")

        lazy_mod = import_lazy("nonexistent")
        lazy_attr = lazy_mod.function

        with pytest.raises(ImportError, match="Failed to lazily import nonexistent: Module not found"):
            lazy_attr()

        assert repr(lazy_mod) == "<lazy nonexistent loaded=True failed=True>"
        assert repr(lazy_attr) == "<lazyattr nonexistent.function loaded=True failed=True>"


def test_missing_attr() -> None:
    mock_module = ModuleType("mock")

    with patch("importlib.import_module") as mock_import_module:
        mock_import_module.return_value = mock_module

        lazy_mod = import_lazy("mock")
        lazy_attr = lazy_mod.function

        with pytest.raises(AttributeError, match="module 'mock' has no attribute 'function'"):
            lazy_attr()

        assert repr(lazy_mod) == "<lazy mock loaded=True failed=False>"
        assert repr(lazy_attr) == "<lazyattr mock.function loaded=True failed=True>"



from __future__ import annotations

import gzip
from textwrap import dedent
from typing import TYPE_CHECKING

from dissect.target.filesystems.ntds import NtdsFilesystem
from dissect.target.target import Target
from dissect.target.tools.shell import TargetCli
from dissect.target.tools.utils import fs
from tests._utils import absolute_path

if TYPE_CHECKING:
    import pytest


def test_ntds() -> None:
    """Test that the NTDS filesystem can be detected and read correctly."""
    with gzip.open(absolute_path("_data/plugins/os/windows/ad/ntds/goad/ntds.dit.gz"), "rb") as fh:
        assert NtdsFilesystem.detect(fh)

        fs = NtdsFilesystem(fh)

        root = fs.get("/")
        assert root.is_dir()
        assert root.name == ""

        users = fs.get("local/sevenkingdoms/Users")
        assert users.is_dir()
        assert users.name == "Users"
        assert (
            users.open()
            .read()
            .decode()
            .startswith(
                dedent(
                    """
                    DNT: 2050
                    Pdnt: 2043
                    Obj: True
                    RdnType: cn
                    """
                ).strip()
            )
        )

        administrator = fs.get("local/sevenkingdoms/Users/Administrator")
        assert administrator.is_file()
        assert administrator.name == "Administrator"
        assert (
            administrator.open()
            .read()
            .decode()
            .startswith(
                dedent(
                    """
                    DNT: 3802
                    Pdnt: 2050
                    Obj: True
                    RdnType: cn
                    """
                ).strip()
            )
        )


def test_shell(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that the shell can be used to read files and directories from the NTDS filesystem."""
    # disable colorful output in `target-shell`
    monkeypatch.setattr(fs, "LS_COLORS", {})

    with gzip.open(absolute_path("_data/plugins/os/windows/ad/ntds/goad/ntds.dit.gz"), "rb") as fh:
        t = Target()
        t.fs.mount("/", NtdsFilesystem(fh))

        cli = TargetCli(t)

        cli.onecmd("cat local/sevenkingdoms/Users")
        captured = capsys.readouterr()
        assert captured.out.startswith("DNT: 2050")

        cli.onecmd("cat local/sevenkingdoms/Users/Administrator")
        captured = capsys.readouterr()
        assert captured.out.startswith("DNT: 3802")



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.unix.linux.suse.zypper import ZypperPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    "test_file",
    [
        "history",
        "history.1.gz",
        "history.1.bz2",
    ],
)
def test_zypper_logs(target_unix: Target, fs_unix: VirtualFilesystem, test_file: str) -> None:
    tz = timezone.utc
    data_file = absolute_path(f"_data/plugins/os/unix/linux/suse/zypp/{test_file}")
    fs_unix.map_file(f"/var/log/zypp/{test_file}", data_file)
    target_unix.add_plugin(ZypperPlugin)

    results = list(target_unix.zypper.logs())
    assert len(results) == 61

    for record in results:
        assert record.package_manager == "zypper"

    assert results[0].ts == datetime(2022, 12, 16, 12, 56, 23, tzinfo=tz)
    assert results[0].operation == "other"
    assert results[0].package_name is None
    assert results[0].command == "zypper install unzip"
    assert results[0].requested_by_user == "root"

    assert results[-1].ts == datetime(2022, 12, 16, 13, 2, 44, tzinfo=tz)
    assert results[-1].operation == "install"
    assert results[-1].package_name == "yast2-4.5.20-1.1:x86_64"
    assert results[-1].command is None
    assert results[-1].requested_by_user is None



from __future__ import annotations

import fnmatch
import re
from typing import TYPE_CHECKING

from dissect.target.helpers import configutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath

UnixConfigTreeRecord = TargetRecordDescriptor(
    "unix/config",
    [
        ("path", "source"),
        ("path", "path"),
        ("string", "key"),
        ("string[]", "value"),
    ],
)


class EtcTree(Plugin):
    """Unix etc configuration tree plugin."""

    __namespace__ = "etc"

    def check_compatible(self) -> None:
        return None

    def _sub(
        self, items: configutil.ConfigurationParser, entry: TargetPath, orig_path: TargetPath, pattern: str
    ) -> Iterator[UnixConfigTreeRecord]:
        index = 0

        for raw_key, value in items.items():
            key = re.sub(r"[\n\r\t/]", "", raw_key)
            path = entry / key

            if isinstance(value, dict):
                yield from self._sub(value, path, orig_path, pattern)
                continue

            if not isinstance(value, list):
                value = [str(value)]

            if fnmatch.fnmatch(path, pattern):
                config_path = entry.relative_to(orig_path)
                data = {
                    "_target": self.target,
                    "source": orig_path,
                    "path": "/" / config_path,
                    "key": key,
                    "value": value,
                }
                if value == [""]:
                    data["key"] = index
                    data["value"] = [key]
                    index += 1

                yield UnixConfigTreeRecord(**data)

    @export(record=UnixConfigTreeRecord)
    @arg("--glob", dest="pattern", default="*", help="glob-style pattern to search for")
    @arg("--root", dest="root", default="/etc", help="path to use as root for search")
    @arg("--unknowns", dest="unknowns", action="store_true", help="return unknown / unparsable objects")
    def etc(self, pattern: str, root: str, unknowns: bool) -> Iterator[UnixConfigTreeRecord]:
        """This plugin yields configuration information from the etc directory in key value pairs.

        Args:
            pattern: What glob-style pattern to search for.
            root: Path to use as root for searching.
            unknowns: Whether to also yield entries for unknown entries.

        Yields UnixConfigTreeRecord with the following fields:

        .. code-block:: text

            source (path): The path on the target used for parsing.
            path (path): The path inside the configuration file that is being used.
            key (string): The configuration key returned by parsing.
            value (string[]): The configuration value belonging to the key.
        """

        for entry, _, items in self.target.fs.walk(root):
            for item in items:
                path = self.target.fs.path(entry) / item
                try:
                    config_object = configutil.parse(self.target.fs.path(path))

                    if not unknowns and isinstance(config_object, (configutil.Txt, configutil.Bin)):
                        # We don't have a specific config parser for this file,
                        # Ignore those config paths to get proper records
                        continue
                    yield from self._sub(config_object, path, orig_path=path, pattern=pattern)
                except Exception:
                    self.target.log.warning("Could not parse file: %s", path)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

CmdlineRecord = TargetRecordDescriptor(
    "linux/proc/cmdline",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("varint", "pid"),
        ("string", "state"),
        ("string", "cmdline"),
    ],
)


class CmdlinePlugin(Plugin):
    """Linux volatile proc commandline plugin."""

    def check_compatible(self) -> None:
        if not self.target.has_function("proc"):
            raise UnsupportedPluginError("proc filesystem not available")

    @export(record=CmdlineRecord)
    def cmdline(self) -> Iterator[CmdlineRecord]:
        """Return the complete command line for all processes.

        If, after an execve(2), the process modifies its argv strings, those changes will show up here. This is not the
        same thing as modifying the argv array.

        Think of this output as the command line that the process wants you to see.

        Yields CmdlineRecord with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The starttime of the process.
            name (string): The name of the process.
            pid (int): The process ID of the process.
            cmdline (string): The complete commandline of the process.
        """

        for process in self.target.proc.processes():
            yield CmdlineRecord(
                ts=process.starttime,
                name=process.name,
                pid=process.pid,
                state=process.state,
                cmdline=process.cmdline,
                _target=self.target,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

AppxDebugKeyRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/registry/appxdebug/key",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "debug_info"),
    ],
)


class AppxDebugKeysPlugin(Plugin):
    """Plugin that iterates various AppX debug key locations."""

    REGKEY_GLOBS = (
        # The first glob are the AppX package names
        # The "(Default Value)" contains the debugger command
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\PackagedAppXDebug\\*",
        # The first glob are the AppX package names
        # The second glob are the AppX package components
        # The "DebugPath" value contains the debugger command
        "HKEY_CURRENT_USER\\Software\\Classes\\ActivatableClasses\\Package\\*\\DebugInformation\\*",
    )

    def _walk(self, key: RegistryKey) -> Iterator[AppxDebugKeyRecord]:
        user = self.target.registry.get_user(key)

        values = key.values()
        subkeys = key.subkeys()

        if not values and not subkeys:
            yield AppxDebugKeyRecord(
                ts=key.ts,
                name=None,
                debug_info=None,
                _target=self.target,
                _key=key,
                _user=user,
            )

        else:
            for value in values:
                yield AppxDebugKeyRecord(
                    ts=key.ts,
                    name=value.name,
                    debug_info=value.value,
                    _target=self.target,
                    _key=key,
                    _user=user,
                )

            for subkey in subkeys:
                yield from self._walk(subkey)

    def _debug_keys(self) -> Iterator[AppxDebugKeyRecord]:
        for regkey_glob in self.REGKEY_GLOBS:
            for key in self.target.registry.glob_ext(regkey_glob):
                yield from self._walk(key)

    def check_compatible(self) -> None:
        try:
            next(self._debug_keys())
        except StopIteration:
            raise UnsupportedPluginError("No registry AppX debug key found")

    @export(record=AppxDebugKeyRecord)
    def appxdebugkeys(self) -> Iterator[AppxDebugKeyRecord]:
        """Iterate various AppX debug key locations. See source for all locations.

        AppX debug keys are registry keys that attach a debugger executable to
        Universal Windows Platform Apps (AppX). This debugger is executed when
        the program is launched and is often leveraged as a persistence
        mechanism.

        References:
            - https://oddvar.moe/2018/09/06/persistence-using-universal-windows-platform-apps-appx/

        Yields AppXDebugKeyRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The registry key last modified timestamp.
            name (string): The AppX debug key name.
            debug_info (string): The AppX debug info.
            regf_hive_path (string): The hive file that contains the registry key.
            regf_key_path (string): The key's full path in the registry.
            username (string): The name of the user this key belongs to.
            user_id (string): The id of the user this key belongs to.
            user_group (string): The group of the user this key belongs to.
            user_home (string): The home directory of the user this key belongs to.
        """

        yield from self._debug_keys()



from __future__ import annotations

import gzip
import io

from dissect.target import container
from dissect.target.containers.vmdk import VmdkContainer
from dissect.target.filesystem import VirtualFilesystem
from tests._utils import absolute_path


def test_vmdk_container() -> None:
    """Test that VMDK containers are properly opened.

    Generated with::

        VBoxManage createmedium disk --filename "./small.vmdk" --size 2 --format=VMDK
    """
    path = absolute_path("_data/containers/vmdk/small.vmdk.gz")
    gz_file = gzip.GzipFile(path)
    fh = container.open(gz_file)
    assert isinstance(fh, VmdkContainer)
    a = fh.read(20)
    assert a == b"\x00" * 20
    assert fh.tell() == 20
    fh.seek(0, whence=io.SEEK_END)
    assert fh.tell() == 2097152
    fh.close()
    gz_file.close()


def test_vmdk_detect_path() -> None:
    """Test that VMDK containers are properly opened, when using the extension based matching."""
    vfs = VirtualFilesystem()
    vfs.map_file("small.vmdk", absolute_path("_data/containers/vmdk/small.vmdk.gz"), compression="gzip")
    fh = container.open(vfs.path("small.vmdk"))
    assert isinstance(fh, VmdkContainer)
    a = fh.read(20)
    assert a == b"\x00" * 20
    assert fh.tell() == 20
    fh.seek(0, whence=io.SEEK_END)
    assert fh.tell() == 2097152



from __future__ import annotations

import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

# Regex to validate RustDesk loglines
RE_LOG_LINE = re.compile(r"\[(.*?)\] (\w+) \[(.*?)\] (.*)")


class RustdeskPlugin(RemoteAccessPlugin):
    """Rustdesk plugin."""

    __namespace__ = "rustdesk"

    RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/rustdesk/log", GENERIC_LOG_RECORD_FIELDS
    )

    # Rustdesk logs when installed as a service/server
    SERVER_GLOBS = (
        # Windows >= Windows 7
        "%windir%/ServiceProfiles/LocalService/AppData/Roaming/RustDesk/log/server/*.log",
        # Linux
        "var/log/rustdesk-server/*.log",
    )

    # User specific Rustdesk logs
    USER_GLOBS = (
        # Windows
        "AppData/Roaming/Rustdesk/log/*.log",
        # Linux
        ".local/share/logs/RustDesk/server/*.log",
        # Android
        "storage/emulated/0/RustDesk/logs/*.log",
        # Mac
        "Library/Logs/RustDesk/*.log",
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.log_files: set[tuple[TargetPath, UserDetails | None]] = set()

        # Service globs
        for log_glob in self.SERVER_GLOBS:
            base, _, glob = log_glob.partition("*")
            for log_file in self.target.resolve(base).glob(f"*{glob}"):
                self.log_files.add((log_file, None))

        # User globs
        for user_details in self.target.user_details.all_with_home():
            for log_glob in self.USER_GLOBS:
                for log_file in user_details.home_path.glob(log_glob):
                    self.log_files.add((log_file, user_details.user))

    def check_compatible(self) -> None:
        if not self.log_files:
            raise UnsupportedPluginError("No Rustdesk log files found on target")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Parse RustDesk log files.

        Rustdesk is a remote desktop application that can be used to get (persistent) access to a machine.
        The project is open source and can be found at: https://github.com/rustdesk/rustdesk/

        The log files are stored in different locations, based on the Target OS and client type.
        Unlike Anydesk, Rustdesk does carry a time zone designator (TZD).

        References:
            - https://rustdesk.com/docs/en/self-host/rustdesk-server-pro/faq
            - https://www.reddit.com/r/rustdesk/comments/1072zst/going_to_need_to_know_where_the_client_installer/
            - https://github.com/IRB0T/Remote-Access-Tools---4N6/blob/main/RustDesk/README.md
            - https://github.com/rustdesk/rustdesk/wiki/FAQ#access-logs
        """
        for log_file, user in self.log_files:
            for line in log_file.open("rt", errors="backslashreplace"):
                if line := line.strip():
                    try:
                        if not (match := RE_LOG_LINE.match(line)):
                            raise ValueError("Line does not match expected format")  # noqa: TRY301

                        ts, level, source, message = match.groups()

                        timestamp = datetime.strptime(ts, "%Y-%m-%d %H:%M:%S.%f %z")
                        message = re.sub(r"\s\s+", " ", f"{level} {source} {message}")

                        yield self.RemoteAccessLogRecord(
                            ts=timestamp,
                            message=message,
                            source=log_file,
                            _target=self.target,
                            _user=user,
                        )
                    except ValueError as e:
                        self.target.log.warning("Could not parse log line in file %s: '%s'", log_file, line)
                        self.target.log.debug("", exc_info=e)



from __future__ import annotations

import operator
import struct
from functools import cached_property
from typing import TYPE_CHECKING, Any
from uuid import UUID

from dissect.target.exceptions import RegistryError, RegistryValueNotFoundError
from dissect.target.helpers.record import WindowsUserRecord
from dissect.target.plugin import OperatingSystem, OSPlugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.plugins.os.windows.sam import SamRecord
    from dissect.target.target import Target

ARCH_MAP = {
    "x86": 32,
    "IA64": 64,
    "ARM64": 64,
    "EM64T": 64,
    "AMD64": 64,
}


class WindowsPlugin(OSPlugin):
    CURRENT_VERSION_KEY = "HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion"

    def __init__(self, target: Target):
        super().__init__(target)

        # Just run this here for now
        self.add_mounts()

        target.props["sysvol_drive"] = next(
            (mnt for mnt, fs in target.fs.mounts.items() if fs is target.fs.mounts.get("sysvol") and mnt != "sysvol"),
            None,
        )

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/windows/system32") or fs.exists("/winnt"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.fs.case_sensitive = False
        target.fs.alt_separator = "\\"
        target.fs.mount("sysvol", sysvol)

        if not sysvol.exists("boot/BCD"):
            for fs in target.filesystems:
                if fs.exists("boot/BCD") or fs.exists("EFI/Microsoft/Boot/BCD"):
                    target.fs.mount("efi", fs)

        return cls(target)

    def add_mounts(self) -> None:
        self.target.props["mounts_added"] = True
        try:
            for key in self.target.registry.keys("HKLM\\System\\MountedDevices"):
                for entry in key.values():
                    name, value = entry.name, entry.value
                    if not name.lower().startswith("\\"):
                        continue

                    p = name.lower()[1:].split("\\")
                    if p[0] == "dosdevices":
                        drive = p[1]

                        if value.startswith(b"DMIO:ID:"):
                            guid = str(UUID(bytes_le=value[8:]))

                            for volume in self.target.volumes:
                                if volume.guid == guid and volume.fs:
                                    self.target.fs.mount(drive, volume.fs)

                        elif len(value) == 12:
                            serial, offset = struct.unpack("<IQ", value)
                            for disk in self.target.disks:
                                if disk.vs and disk.vs.serial == serial:
                                    for volume in disk.vs.volumes:
                                        if volume.offset == offset and volume.fs:
                                            self.target.fs.mount(drive, volume.fs)
                                            break
        except Exception as e:
            self.target.log.warning("Failed to map drive letters")
            self.target.log.debug("", exc_info=e)

        sysvol_drive = self.target.fs.mounts.get("sysvol")
        if not sysvol_drive:
            self.target.log.warning("No sysvol drive found")
        elif operator.countOf(self.target.fs.mounts.values(), sysvol_drive) == 1:
            # Fallback mount the sysvol to C: if we didn't manage to mount it to any other drive letter
            if "c:" not in self.target.fs.mounts:
                self.target.log.debug("Unable to determine drive letter of sysvol, falling back to C:")
                self.target.fs.mount("c:", sysvol_drive)
            else:
                self.target.log.warning("Unknown drive letter for sysvol")

    @export(property=True)
    def hostname(self) -> str | None:
        key = "HKLM\\SYSTEM\\ControlSet001\\Control\\Computername\\Computername"
        try:
            return self.target.registry.value(key, "Computername").value
        except RegistryError:
            return None

    @export(property=True)
    def ips(self) -> list[str]:
        return list(set(map(str, self.target.network.ips())))

    def _get_version_reg_value(self, value_name: str) -> Any:
        try:
            value = self.target.registry.value(self.CURRENT_VERSION_KEY, value_name).value
        except RegistryError:
            value = None

        return value

    def _legacy_current_version(self) -> str | None:
        """Returns the NT version as used up to and including NT 6.3.

        This corresponds with Windows 8 / Windows 2012 Server.

        Returns:
            The string value of the NT version number or ``None`` if the
            CurrentVersion sub key is not present or any other registry error
            occurred.
        """
        return self._get_version_reg_value("CurrentVersion")

    def _major_version(self) -> int | None:
        """Return the NT major version number (starting from NT 10.0 / Windows 10).

        Returns:
            The integer value of the NT major version number or ``None`` if the
            CurrentMajorVersionNumber sub key is not present or any other
            registry error occurred.
        """
        return self._get_version_reg_value("CurrentMajorVersionNumber")

    def _minor_version(self) -> int | None:
        """Return the NT minor version number (starting from NT 10.0 / Windows 10).

        Returns:
            The integer value of the NT minor version number or ``None`` if the
            CurrentMinorVersionNumber sub key is not present or any other
            registry error occurred.
        """
        return self._get_version_reg_value("CurrentMinorVersionNumber")

    def _nt_version(self) -> int | None:
        """Return the Windows NT version in x.y format.

        For systems up to and including NT 6.3 (Win 8 / Win 2012 Server) this
        will be the value of the CurrentVersion sub key.

        For systems starting from NT 10.0 (Win 10) this will be a value
        constructed from the combination of the CurrentMajorVersionNumber and
        CurrentMinorVersionNumber sub keys.

        Returns:
            The string value of the NT version or ``None`` if the any of the
            sub keys are not present or any other registry error occurred.
        """
        version = None

        major_version = self._major_version()
        if major_version is not None:
            minor_version = self._minor_version()
            if minor_version is None:
                minor_version = ""
            version = f"{major_version}.{minor_version}"
        else:
            version = self._legacy_current_version()

        return version

    @export(property=True)
    def version(self) -> str | None:
        """Return a string representation of the Windows version of the target.

        For Windows versions before Windows 10 this looks like::

            <ProductName> (NT <CurrentVersion>) <CurrentBuildNumber>.<UBR> <CSDVersion>

        For Windows versions since Windows 10 this looks like::

            <ProductName> (NT <CurrentMajorVersionNumber>.<CurrentMinorVersionNumber>) <CurrentBuildNumber>.<UBR> <CSDVersion>

        Where the registry values used are between ``<...>``.

        Note that the ``<UBR>`` and ``<CSDVersion>`` may or may not be available,
        depending on whether updates and service packs are installed.

        Note also that we don't show the "version" (aka FeatureRelease) as
        shown by WinVer.exe, which uses the registry values:

        ``<ReleaseId>``: Windows up to Windows 10 ReleaseId <= 2004

        ``<DisplayVersion>``: from Windows 10 ReleaseId >= 2009
                              (DisplayVersion = 20H2 in this case)

        Returns:
            If any one of the registry values used in the version string can be
            found in the registry, a string is returned as described above.
            All values that can not be found and should be present are replaced
            with ``<UNKNOWN value_name>``.
            If none of the values can be found, ``None`` is returned.
        """  # noqa: E501
        # https://www.vcloudinfo.com/2020/12/how-to-decode-windows-version-numbers.html

        def _part_str(parts: dict[str, Any], name: str) -> str:
            value = parts.get(name)
            return f"<Unknown {name}>" if value is None else str(value)

        version_parts = {}
        version_parts["ProductName"] = self._get_version_reg_value("ProductName")
        version_parts["CurrentVersion"] = self._nt_version()
        # Note that the CurrentBuild key value also exists, often with the
        # same value, but is apparently deprecated.
        version_parts["CurrentBuildNumber"] = self._get_version_reg_value("CurrentBuildNumber")
        # Update Build Revision (seems to be present from NT 6.1 onwards).
        version_parts["UBR"] = self._get_version_reg_value("UBR")
        # Service Pack Version, when installed using the normal method (may not be present).
        version_parts["CSDVersion"] = self._get_version_reg_value("CSDVersion")

        version_string = None
        if any(value is not None for value in version_parts.values()):
            version = []

            nt_version = _part_str(version_parts, "CurrentVersion")
            build_version = _part_str(version_parts, "CurrentBuildNumber")
            prodcut_name = _part_str(version_parts, "ProductName")

            # CurrentBuildNumber >= 22000 on NT 10.0 indicates Windows 11.
            # https://learn.microsoft.com/en-us/windows/release-health/windows11-release-information
            try:
                if nt_version == "10.0" and int(build_version) >= 22_000:
                    prodcut_name = prodcut_name.replace("Windows 10", "Windows 11")
            except ValueError:
                pass

            version.append(prodcut_name)
            version.append(f"(NT {nt_version})")

            ubr = version_parts["UBR"]
            if ubr:
                build_version = f"{build_version}.{ubr}"
            version.append(build_version)

            csd_version = version_parts["CSDVersion"]
            if csd_version is not None:
                version.append(f"{csd_version}")

            version_string = " ".join(version)

        return version_string

    @export(property=True)
    def architecture(self) -> str | None:
        """Returns a target triple containing the architecture and bitness of the system.

        Returns:
            Target triple string.
        """

        key = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"

        try:
            arch = self.target.registry.key(key).value("PROCESSOR_ARCHITECTURE").value
            bits = ARCH_MAP.get(arch)

            if bits == 64:
                return f"{arch}-win{bits}".lower()
            return f"{arch}_{bits}-win{bits}".lower()
        except RegistryError:
            pass

    @cached_property
    def _sam_by_sid(self) -> dict[str, SamRecord]:
        if not (machine_sid := next(self.target.machine_sid(), None)):
            return {}

        sam_users: dict[str, SamRecord] = {}
        try:
            for sam_record in self.target.sam():
                # Compose SID from domain_sid and RID
                sam_users[f"{machine_sid.sid}-{sam_record.rid}"] = sam_record
        except Exception as e:
            self.target.log.warning("Could not read SAM records")
            self.target.log.debug("", exc_info=e)

        return sam_users

    @export(record=WindowsUserRecord)
    def users(self) -> Iterator[WindowsUserRecord]:
        # Be aware that this function can never do anything which needs user
        # registry hives. Initializing those hives will need this function,
        # which will then cause a recursion.

        key = "HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList"
        sids = set()
        for k in self.target.registry.keys(key):
            for subkey in k.subkeys():
                sid = subkey.name

                if sid in sids:
                    continue

                sids.add(sid)
                home = None
                name = None
                try:
                    profile_image_path = subkey.value("ProfileImagePath")
                except RegistryValueNotFoundError:
                    pass
                else:
                    home = profile_image_path.value
                    # Use SAM username if available
                    name = self._sam_by_sid[sid].username if sid in self._sam_by_sid else home.split("\\")[-1]

                yield WindowsUserRecord(
                    sid=subkey.name,
                    name=name,
                    home=self.target.resolve(home),
                    _target=self.target,
                )

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        yield from (
            (resolved_path, user_criterion)
            for path, user_criterion in [
                ("%windir%/ServiceProfiles/LocalService", ("sid", "S-1-5-19")),
                ("%windir%/ServiceProfiles/NetworkService", ("sid", "S-1-5-20")),
                ("%windir%/System32/config/systemprofile", ("sid", "S-1-5-18")),
                ("%windir%/SysWOW64/config/systemprofile", ("sid", "S-1-5-18")),
            ]
            if (resolved_path := self.target.resolve(path)).exists()
        )

        # Homedirs for users when there is no user profile information available.
        for user_dir in ("sysvol/Users", "sysvol/Documents and Settings"):
            if (resolved := self.target.resolve(user_dir)).exists():
                yield from ((entry, None) for entry in resolved.iterdir() if entry.is_dir())

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.WINDOWS.value



from __future__ import annotations

import re

from dissect.target.helpers.record import TargetRecordDescriptor

DefenderMPLogProcessImageRecord = TargetRecordDescriptor(
    "windows/defender/mplog/processimage",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "process_image_name"),
        ("varint", "pid"),
        ("varint", "total_time"),
        ("varint", "count"),
        ("varint", "max_time"),
        ("string", "max_time_file"),
        ("varint", "estimated_impact"),
    ],
)

DefenderMPLogMinFilUSSRecord = TargetRecordDescriptor(
    "windows/defender/mplog/minfiluss",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("path", "path"),
        ("string", "process"),
        ("string", "status"),
        ("string", "state"),
        ("string", "scan_request"),
        ("string", "file_id"),
        ("string", "reason"),
        ("string", "io_status_block_for_new_file"),
        ("string", "desired_access"),
        ("string", "file_attributes"),
        ("string", "scan_attributes"),
        ("string", "access_state_flags"),
        ("string", "backing_file_info"),
    ],
)

DefenderMPLogMinFilBlockedFileRecord = TargetRecordDescriptor(
    "windows/defender/mplog/blockedfile",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "blocked_file"),
        ("string", "process"),
        ("string", "status"),
        ("string", "state"),
        ("string", "scan_request"),
        ("string", "file_id"),
        ("string", "reason"),
        ("string", "io_status_block_for_new_file"),
        ("string", "desired_access"),
        ("string", "file_attributes"),
        ("string", "scan_attributes"),
        ("string", "access_state_flags"),
        ("string", "backing_file_info"),
    ],
)


DefenderMPLogBMTelemetryRecord = TargetRecordDescriptor(
    "windows/defender/mplog/bmtelemetry",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "guid"),
        ("varint", "signature_id"),
        ("string", "sigsha"),
        ("varint", "threat_level"),
        ("varint", "process_id"),
        ("varint", "process_creation_time"),
        ("varint", "session_id"),
        ("path", "image_path"),
        ("string", "taint_info"),
        ("string", "operations"),
    ],
)

DefenderMPLogEMSRecord = TargetRecordDescriptor(
    "windows/defender/mplog/ems",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "process"),
        ("varint", "pid"),
        ("string", "sigseq"),
        ("varint", "send_memory_scan_report"),
        ("varint", "source_id"),
    ],
)

DefenderMPLogOriginalFileNameRecord = TargetRecordDescriptor(
    "windows/defender/mplog/originalfilename",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "original_file_name"),
        ("path", "full_path"),
        ("string", "hr"),
    ],
)

DefenderMPLogExclusionRecord = TargetRecordDescriptor(
    "windows/defender/mplog/exclusion",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("path", "full_path_with_drive_letter"),
        ("path", "full_path_with_device_path"),
    ],
)

DefenderMPLogLowfiRecord = TargetRecordDescriptor(
    "windows/defender/mplog/lowfi",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("command", "lowfi"),
    ],
)

DefenderMPLogDetectionAddRecord = TargetRecordDescriptor(
    "windows/defender/mplog/detectionadd",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "detection"),
    ],
)


DefenderMPLogThreatRecord = TargetRecordDescriptor(
    "windows/defender/mplog/threat",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("command", "threat"),
    ],
)

DefenderMPLogDetectionEventRecord = TargetRecordDescriptor(
    "windows/defender/mplog/detectionevent",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "threat_type"),
        ("command", "command"),
    ],
)

DefenderMPLogResourceScanRecord = TargetRecordDescriptor(
    "windows/defender/mplog/resourcescan",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "scan_id"),
        ("varint", "scan_source"),
        ("datetime", "start_time"),
        ("datetime", "end_time"),
        ("string", "resource_schema"),
        ("path", "resource_path"),
        ("varint", "result_count"),
        ("string[]", "threats"),
        ("path[]", "resources"),
    ],
)

DefenderMPLogThreatActionRecord = TargetRecordDescriptor(
    "windows/defender/mplog/threataction",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string[]", "threats"),
        ("path[]", "resources"),
        ("string[]", "actions"),
    ],
)

DefenderMPLogRTPRecord = TargetRecordDescriptor(
    "windows/defender/mplog/rtp_log",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("datetime", "last_perf"),
        ("datetime", "first_rtp_scan"),
        ("string", "plugin_states"),
        ("path[]", "process_exclusions"),
        ("path[]", "path_exclusions"),
        ("string[]", "ext_exclusions"),
    ],
)

DEFENDER_MPLOG_TS_PATTERN = (
    r"(?P<ts>[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}"
    r"\.[0-9]{3}(?:Z|[+-][0-9]{2}:[0-9]{2})?)"
)

# Loosely based on https://github.com/Intrinsec/mplog_parser but feel free to add patterns

DEFENDER_MPLOG_PATTERNS = [
    # Process Image
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s
                ProcessImageName:\s(?P<process_image_name>.*),\s
                Pid:\s(?P<pid>\d*),\s
                TotalTime:\s(?P<total_time>\d*),\s
                Count:\s(?P<count>\d*),\s
                MaxTime:\s(?P<max_time>\d*),\s
                MaxTimeFile:\s(?P<max_time_file>.*),\s
                EstimatedImpact:\s(?P<estimated_impact>\d*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogProcessImageRecord,
    ),
    # Mini-filter Unsuccessful scan status
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s
                \[Mini-filter\]\s
                (Unsuccessful\sscan\sstatus)[^:]*:\s(?P<path>.+)\s
                Process:\s(?P<process>.+),\s
                Status:\s(?P<status>.+),\s
                State:\s(?P<state>.+),\s
                ScanRequest\s(?P<scan_request>.+),\s
                FileId:\s(?P<file_id>.+),\s
                Reason:\s(?P<reason>.+),\s
                IoStatusBlockForNewFile:\s(?P<io_status_block_for_new_file>.+),\s
                DesiredAccess:(?P<desired_access>.+),\s
                FileAttributes:(?P<file_attributes>.+),\s
                ScanAttributes:(?P<scan_attributes>.+),\s
                AccessStateFlags:(?P<access_state_flags>.+),\s
                BackingFileInfo:\s(?P<backing_file_info>.+)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogMinFilUSSRecord,
    ),
    # EMS Scan
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                process:\s(?P<process>\w*)\s
                pid:\s(?P<pid>\d*),\s
                sigseq:\s(?P<sigseq>\w*),\s
                sendMemoryScanReport:\s(?P<send_memory_scan_report>\d*),\s
                source:\s(?P<source_id>\d*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogEMSRecord,
    ),
    # Original filename
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                original\sfile\sname\s\"(?P<original_file_name>.*)\"\sfor\s\"(?P<full_path>.*)\",\shr=(?P<hr>\w*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogOriginalFileNameRecord,
    ),
    # Mini-filter Blocked file
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                \[Mini-filter\]\s
                Blocked\sfile:\s(?P<blocked_file>.+)\s
                Process:\s(?P<process>.+),\s
                Status:\s(?P<status>.+),\s
                State:\s(?P<state>.+),\s
                ScanRequest\s(?P<scan_request>.+),\s
                FileId:\s(?P<file_id>.+),\s
                Reason:\s(?P<reason>.+),\s
                IoStatusBlockForNewFile:\s(?P<io_status_block_for_new_file>.+),\s
                DesiredAccess:(?P<desired_access>.+),\s
                FileAttributes:(?P<file_attributes>.+),\s
                ScanAttributes:(?P<scan_attributes>.+),\s
                AccessStateFlags:(?P<access_state_flags>.+),\s
                BackingFileInfo:\s(?P<backing_file_info>.+)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogMinFilBlockedFileRecord,
    ),
    # Exclusion
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s
                \[Exclusion\]\s(?P<full_path_with_drive_letter>.+)\s->\s(?P<full_path_with_device_path>.+)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogExclusionRecord,
    ),
    # Lowfi
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                lowfi:\s(?P<lowfi>.+)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogLowfiRecord,
    ),
    # Detection add
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                DETECTION_ADD\S*\s(?P<detection>.*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogDetectionAddRecord,
    ),
    # Threat
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                threat:\s(?P<threat>.*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogThreatRecord,
    ),
    # Detection event
    (
        re.compile(
            rf"""
                {DEFENDER_MPLOG_TS_PATTERN}\s.*
                DETECTIONEVENT\sMPSOURCE_\S+\sHackTool:(?P<threat_type>.*)\sfile:(?P<command>.*)
            """,
            re.VERBOSE,
        ),
        DefenderMPLogDetectionEventRecord,
    ),
]


DEFENDER_MPLOG_BLOCK_PATTERNS = [
    (
        re.compile(r"Begin Resource Scan"),
        re.compile(r"End Scan"),
        re.compile(
            r"""
                Begin\sResource\sScan.*\n
                Scan\sID:(?P<scan_id>[^\n]+)\n
                Scan\sSource:(?P<scan_source>\d+)\n
                Start\sTime:(?P<start_time>[0-9\-\:\s]*)\n
                End\sTime:(?P<end_time>[0-9\-\:\s]*)\n
                .*
                Resource\sSchema:(?P<resource_schema>[^\n]+)\n
                Resource\sPath:(?P<resource_path>[^\n]+)\n
                Result\sCount:(?P<result_count>\d+)\n
                (?P<rest>.*)\n
                End\sScan
            """,
            re.VERBOSE | re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogResourceScanRecord,
    ),
    # Threat actions
    (
        re.compile(r"Beginning threat actions"),
        re.compile(r"Finished threat actions"),
        re.compile(
            r"""
            Beginning\sthreat\sactions\n
            Start\stime:(?P<ts>[0-9\-\:\s]*)\n
            (?P<rest>.*)\n
            Finished\sthreat\sactions
            """,
            re.VERBOSE | re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogThreatActionRecord,
    ),
    # RTP
    (
        re.compile(r"\*\*RTP Perf Log\*\*"),
        re.compile(r"\*\*END RTP Perf Log\*\*"),
        re.compile(
            r"""
            \*+RTP\sPerf\sLog\*+\n
            RTP\sStart:(?P<ts>.*)\n
            Last\sPerf:(?P<last_perf>.*)\n
            First\sRTP\sScan:(?P<first_rtp_scan>.*)\n
            Plugin\sStates:(?P<plugin_states>.*)\n
            Process\sExclusions:\n(?P<process_exclusions>.*)
            Path\sExclusions:\n(?P<path_exclusions>.*)
            Ext\sExclusions:\n(?P<ext_exclusions>.*)
            Worker\sThreads
            """,
            re.VERBOSE | re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogRTPRecord,
    ),
    # BM Telemetry (block)
    (
        re.compile(r"BEGIN BM telemetry"),
        re.compile(r"END BM telemetry"),
        re.compile(
            r"""
                BEGIN\sBM\stelemetry\n
                (GUID):(?P<guid>.+)\n
                (SignatureID):(?P<signature_id>.+)\n
                (SigSha):(?P<sigsha>.+)\n
                (ThreatLevel):(?P<threat_level>.+)\n
                (ProcessID):(?P<process_id>.+)\n
                (ProcessCreationTime):(?P<process_creation_time>.+)\n
                (SessionID):(?P<session_id>.+)\n
                (CreationTime):(?P<ts>.+)\n
                (ImagePath):(?P<image_path>.+)\n
                (Taint\sInfo):(?P<taint_info>.+)\n
                (Operations):(?P<operations>.+)\n
                END\sBM\stelemetry
            """,
            re.VERBOSE,
        ),
        DefenderMPLogBMTelemetryRecord,
    ),
]

DEFENDER_MPLOG_LINE = re.compile(r"^\s+(.*)$", re.MULTILINE)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.tools.utils.logging import configure_logging

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture(autouse=True)
def prevent_logging_setup(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    def noop(*args, **kwargs) -> None:
        pass

    with monkeypatch.context() as m:
        m.setattr(configure_logging, "__code__", noop.__code__)
        yield



from __future__ import annotations

import re
from collections import OrderedDict
from configparser import ConfigParser
from functools import partial
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import OperatingSystem, Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

WireGuardInterfaceRecord = TargetRecordDescriptor(
    "application/vpn/wireguard/interface",
    [
        ("string", "name"),  # basename of .conf file if unset
        ("net.ipaddress", "address"),
        ("string", "private_key"),
        ("varint", "listen_port"),
        ("string", "fw_mark"),
        ("net.ipaddress[]", "dns"),
        ("string", "table"),
        ("varint", "mtu"),
        ("string", "preup"),
        ("string", "postup"),
        ("string", "predown"),
        ("string", "postdown"),
        ("path", "source"),
    ],
)

WireGuardPeerRecord = TargetRecordDescriptor(
    "application/vpn/wireguard/peer",
    [
        ("string", "name"),
        ("string", "public_key"),
        ("string", "pre_shared_key"),
        ("net.ipnetwork[]", "allowed_ips"),
        ("string", "endpoint"),
        ("varint", "persistent_keep_alive"),
        ("path", "source"),
    ],
)


class WireGuardPlugin(Plugin):
    """WireGuard configuration parser.

    References:
        - https://manpages.debian.org/testing/wireguard-tools/wg.8.en.html#CONFIGURATION_FILE_FORMAT
        - https://github.com/pirate/wireguard-docs
        - https://wiresock.net/documentation/wireguard/config.html
    """

    __namespace__ = "wireguard"

    # TODO: NetworkManager uses a different stanza format
    #       "/etc/NetworkManager/system-connections/Wireguard*",
    # TODO: systemd uses a different stanza format
    #       "/etc/systemd/network/wg*.netdev",
    #       "/etc/systemd/network/*wg*.netdev",
    # TODO: other locations such as $HOME/.config/wireguard
    # TODO: parse native network manager formats from MacOS, Ubuntu and Windows.

    CONFIG_GLOBS = (
        # Linux
        "/etc/wireguard/*.conf",
        # MacOS
        "/usr/local/etc/wireguard/*.conf",
        "/opt/homebrew/etc/wireguard/*.conf",
        # Windows
        "C:\\Windows\\System32\\config\\systemprofile\\AppData\\Local\\WireGuard\\Configurations\\*.dpapi",
        "C:\\Program Files\\WireGuard\\Data\\Configurations\\*.dpapi",
    )

    TUNNEL_NAME_RE = re.compile(r"\.(conf(\.dpapi)?|netdev)$")

    def __init__(self, target: Target):
        super().__init__(target)
        self.configs: list[Path] = []
        for path in self.CONFIG_GLOBS:
            self.configs.extend(self.target.fs.path("/").glob(path.lstrip("/")))

    def check_compatible(self) -> None:
        if not self.configs:
            raise UnsupportedPluginError("No Wireguard configuration files found")

    @export(record=[WireGuardInterfaceRecord, WireGuardPeerRecord])
    def config(self) -> Iterator[WireGuardInterfaceRecord | WireGuardPeerRecord]:
        """Parses interface config files from wireguard installations."""

        for config_path in self.configs:
            if self.target.os == OperatingSystem.WINDOWS and config_path.suffix == ".dpapi":
                try:
                    config_buf = self.target.dpapi.decrypt_system_blob(config_path.read_bytes())
                    config_buf = config_buf.decode()
                except ValueError:
                    self.target.log.warning("Failed to decrypt WireGuard configuration at %s", config_path)
                    continue
            else:
                config_buf = config_path.read_text()

            config = _parse_config(config_buf)

            # Set up an iterator to go through all the sections and pre-set the fallback
            config_iterator = ((section, partial(config.get, section, fallback=None)) for section in config.sections())

            for section, config_dict in config_iterator:
                if "Interface" in section:
                    if address := config_dict("Address"):
                        address = address.split("/")[0]

                    yield WireGuardInterfaceRecord(
                        name=config_path.stem,
                        address=address,
                        listen_port=config_dict("ListenPort"),
                        private_key=config_dict("PrivateKey"),
                        fw_mark=config_dict("FwMark"),
                        dns=[ip.strip() for ip in config_dict("DNS").split(",")] if config_dict("DNS") else None,
                        table=config_dict("Table"),
                        mtu=config_dict("MTU"),
                        preup=config_dict("PreUp"),
                        postup=config_dict("PostUp"),
                        predown=config_dict("PreDown"),
                        postdown=config_dict("PostDown"),
                        source=config_path,
                        _target=self.target,
                    )

                if "Peer" in section:
                    if allowed_ips := config_dict("AllowedIPs"):
                        allowed_ips = [value.strip() for value in allowed_ips.split(",")]

                    yield WireGuardPeerRecord(
                        name=config_dict("Name"),
                        public_key=config_dict("PublicKey"),
                        pre_shared_key=config_dict("PreSharedKey"),
                        allowed_ips=allowed_ips,
                        endpoint=config_dict("Endpoint"),
                        persistent_keep_alive=config_dict("PersistentKeepAlive"),
                        source=config_path,
                        _target=self.target,
                    )


def _parse_config(content: str) -> ConfigParser:
    """Parses WireGuard config ini stanza files using Python's ConfigParser module.

    We create our own MultiDict definition from an ordered dict, since we want
    to allow multiple 'Peer' sections. We prepend section names with numbers;
    'Interface', 'Peer1', 'Peer2', 'Peer2', etc.
    This way we prevent duplicate section keys.
    """

    cp = ConfigParser(defaults=None, dict_type=MultiDict, strict=False)
    # Set to use str so it doesn't do any lower operation on the keys.
    cp.optionxform = str
    cp.read_string(content)
    return cp


class MultiDict(OrderedDict):
    """OrderedDict implementation which allows multiple values for the keys ``Peer`` and ``Interface``."""

    def __init__(self, *args, **kwargs):
        self._unique = 0
        super().__init__(*args, **kwargs)

    def __setitem__(self, key: str, val: str):
        if isinstance(val, dict) and (key in ["Peer", "Interface"]):
            self._unique += 1
            key += str(self._unique)
        super().__setitem__(key, val)



from __future__ import annotations

import tempfile
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.filesystem.yara import HAS_YARA, YaraPlugin, is_valid_yara
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target

if HAS_YARA:
    import yara

rule_file = absolute_path("_data/plugins/filesystem/yara/rule-dir/rule.yar")
another_rule_file = absolute_path("_data/plugins/filesystem/yara/rule-dir/another.yar")
invalid_rule = absolute_path("_data/plugins/filesystem/yara/rule-dir/invalid.yar")
rule_dir = rule_file.parent


@pytest.fixture
def target_yara(target_default: Target) -> Target:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test_file", BytesIO(b"test string"))
    vfs.map_file_fh("/test/dir/to/test_file", BytesIO(b"test string"))
    vfs.map_file_fh("should_not_hit", BytesIO(b"this is another file."))
    target_default.fs.mount("/", vfs)
    target_default.add_plugin(YaraPlugin)
    return target_default


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
def test_yara_plugin(target_yara: Target) -> None:
    results = list(target_yara.yara(rules=[rule_file]))

    assert len(results) == 2
    assert results[0].path == "/test_file"
    assert results[0].rule == "test_rule_name"
    assert results[0].matches == ["$=test string"]
    assert results[1].tags == ["tag1", "tag2", "tag3"]
    assert results[0].digest.sha1 == "661295c9cbf9d6b2f6428414504a8deed3020641"

    assert results[1].path == "/test/dir/to/test_file"
    assert results[1].rule == "test_rule_name"
    assert results[1].matches == ["$=test string"]
    assert results[1].tags == ["tag1", "tag2", "tag3"]
    assert results[1].digest.sha1 == "661295c9cbf9d6b2f6428414504a8deed3020641"


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
@pytest.mark.parametrize(
    ("rules", "expected_hits", "should_be_valid"),
    [
        (["/does/not/exist"], 0, False),
        ([rule_file, rule_file], 2, True),
        ([rule_file, another_rule_file], 4, True),
        ([rule_dir], 4, False),  # contains invalid.yar
        ([invalid_rule], 0, False),
    ],
)
def test_yara_plugin_invalid_rules(
    target_yara: Target, rules: list[str | Path], expected_hits: int, should_be_valid: bool
) -> None:
    assert is_valid_yara(files={str(file): file for file in rules}) == should_be_valid

    results = list(target_yara.yara(rules=rules, check=True))
    assert len(results) == expected_hits


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
def test_yara_plugin_invalid_rule_warn(target_yara: Target, caplog: pytest.CaptureFixture) -> None:
    results = list(target_yara.yara(rules=[invalid_rule, another_rule_file], check=True))
    assert "invalid.yar contains invalid rule(s)!" in caplog.text
    assert len(results) == 2


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
def test_yara_plugin_compiled_rule(target_yara: Target, tmp_path: str) -> None:
    with tempfile.NamedTemporaryFile(mode="w", dir=tmp_path, delete=False) as tf:
        rules = yara.compile(str(rule_file))
        rules.save(tf.name)
        tf.close()

        results = list(target_yara.yara(rules=[tf.name]))

        assert len(results) == 2

        assert results[0].path == "/test_file"
        assert results[0].rule == "test_rule_name"
        assert results[0].matches == ["$=test string"]
        assert results[0].tags == ["tag1", "tag2", "tag3"]
        assert results[0].namespace == "default"
        assert results[0].digest.md5 == "6f8db599de986fab7a21625b7916589c"
        assert results[0].digest.sha1 == "661295c9cbf9d6b2f6428414504a8deed3020641"
        assert results[0].digest.sha256 == "d5579c46dfcc7f18207013e65b44e4cb4e2c2298f4ac457ba8f82743f31e930b"


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
def test_yara_plugin_multiple_matches(target_default: Target) -> None:
    """Test if we iterate ``yara.StringMatch`` results correctly."""

    vfs = VirtualFilesystem()
    vfs.map_file_fh("/files/first.txt", BytesIO(b"This is a t\x00e\x00s\x00t\x00 file\x00\xaa\xbb\xcc\xdd\xee\xff\x00"))
    vfs.map_file_fh("/files/second.txt", BytesIO(b"This is another test file!"))
    vfs.map_file_fh("/files/third.txt", BytesIO(b"This is some other file!"))
    target_default.fs.mount("/", vfs)
    target_default.add_plugin(YaraPlugin)

    rule = absolute_path("_data/plugins/filesystem/yara/multi.yar")
    results = list(target_default.yara(rules=[str(rule)]))

    assert len(results) == 3

    assert results[0].path == "/files/first.txt"
    assert results[0].matches == ["$a=t\x00e\x00s\x00t\x00", "$b=This", "$c=\\xaa\\xbb\\xcc\\xdd\\xee\\xff"]



from __future__ import annotations

from operator import itemgetter
from typing import TYPE_CHECKING

from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class VyosPlugin(LinuxPlugin):
    def __init__(self, target: Target):
        self.target = target

        versions = []
        for d in self.target.fs.path("/boot").iterdir():
            if d.joinpath("live-rw").exists():
                versions.append((d.name, "live-rw"))
            elif d.joinpath("rw").exists():
                versions.append((d.name, "rw"))

        latest = sorted(versions, key=itemgetter(0))[0]
        self._version, rootpath = latest

        # VyOS does some additional magic with base system files
        layer = target.fs.append_layer()
        layer.map_file_entry("/", target.fs.root.get(f"/boot/{self._version}/{rootpath}"))
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            boot_dir = fs.path("/boot")

            if not boot_dir.is_dir():
                continue

            for d in boot_dir.iterdir():
                if not d.is_dir():
                    continue

                if d.joinpath("live-rw").exists() or d.joinpath("rw").exists():
                    return fs

        return None

    @export(property=True)
    def ips(self) -> list[str] | None:
        return None

    @export(property=True)
    def version(self) -> str:
        return self._version

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.VYOS.value



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

RECORD_NAME = "linux/service"

DEFAULT_ELEMENTS = [
    ("datetime", "ts"),
    ("string", "type"),
    ("string", "name"),
    ("path", "source"),
]

LinuxServiceRecord = TargetRecordDescriptor(RECORD_NAME, DEFAULT_ELEMENTS)


class ServicesPlugin(Plugin):
    """Linux services plugin."""

    SYSTEMD_PATHS = (
        "/etc/systemd/system",
        "/lib/systemd/system",
        "/usr/lib/systemd/system",
    )

    INITD_PATHS = ("/etc/rc.d/init.d", "/etc/init.d")

    def check_compatible(self) -> None:
        if not any(self.target.fs.path(p).exists() for p in self.SYSTEMD_PATHS + self.INITD_PATHS):
            raise UnsupportedPluginError("No supported service directories found")

    @export(record=LinuxServiceRecord)
    def services(self) -> Iterator[LinuxServiceRecord]:
        """Return information about all installed systemd and init.d services.

        References:
            - https://geeksforgeeks.org/what-is-init-d-in-linux-service-management
            - http://0pointer.de/blog/projects/systemd-for-admins-3.html
            - https://www.freedesktop.org/software/systemd/man/latest/systemd.syntax.html
        """

        return chain(self.systemd(), self.initd())

    @internal
    def systemd(self) -> Iterator[LinuxServiceRecord]:
        ignored_suffixes = [".wants", ".requires", ".d"]

        for systemd_path in self.SYSTEMD_PATHS:
            path = self.target.fs.path(systemd_path)
            if not path.exists() or not path.is_dir():
                continue

            for service_file in path.iterdir():
                if should_ignore_file(service_file.name, ignored_suffixes):
                    continue

                config = {}
                types = []

                try:
                    for segment, configuration in configutil.parse(service_file, hint="systemd").items():
                        if not configuration:
                            continue

                        if not isinstance(configuration, list):
                            configuration = [configuration]

                        for nested_conf in configuration:
                            for key, value in nested_conf.items():
                                _key = f"{segment}_{key}".replace("-", "_")

                                if _key not in config:
                                    config.update({_key: value or None})
                                else:
                                    if not isinstance(config[_key], list):
                                        config[_key] = [config[_key]]
                                    config[_key].append(value or None)

                                if isinstance(config[_key], list):
                                    types.append(("string[]", _key))
                                else:
                                    types.append(("string", _key))

                except FileNotFoundError:
                    # The service is registered but the symlink is broken.
                    yield LinuxServiceRecord(
                        ts=service_file.stat(follow_symlinks=False).st_mtime,
                        type="systemd",
                        name=service_file.name,
                        source=service_file,
                        _target=self.target,
                    )
                    continue

                record = TargetRecordDescriptor(RECORD_NAME, DEFAULT_ELEMENTS + types)
                yield record(
                    ts=service_file.stat().st_mtime,
                    type="systemd",
                    name=service_file.name,
                    source=service_file,
                    **config,
                    _target=self.target,
                )

    @internal
    def initd(self) -> Iterator[LinuxServiceRecord]:
        ignored_suffixes = ["README"]

        for initd_path in self.INITD_PATHS:
            path = self.target.fs.path(initd_path)

            if not path.exists():
                continue
            for file_ in path.iterdir():
                if should_ignore_file(file_.name, ignored_suffixes):
                    continue

                yield LinuxServiceRecord(
                    ts=file_.stat().st_mtime,
                    type="initd",
                    name=file_.name,
                    source=file_,
                    _target=self.target,
                )


def should_ignore_file(needle: str, haystack: list) -> bool:
    return needle.endswith(tuple(haystack))



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.xfs import xfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Iterator


log = get_logger(__name__)


class XfsFilesystem(Filesystem):
    __type__ = "xfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.xfs = xfs.XFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        sector = fh.read(512)
        return sector[:4] == b"XFSB"

    def get(self, path: str) -> FilesystemEntry:
        return XfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: xfs.INode | None = None) -> xfs.INode:
        try:
            return self.xfs.get(path, node)
        except xfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except xfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except xfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except xfs.Error as e:
            raise FileNotFoundError(path) from e


class XfsDirEntry(DirEntry):
    fs: XfsFilesystem
    entry: xfs.INode

    def get(self) -> XfsFilesystemEntry:
        return XfsFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class XfsFilesystemEntry(FilesystemEntry):
    fs: XfsFilesystem
    entry: xfs.INode

    def get(self, path: str) -> FilesystemEntry:
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return XfsFilesystemEntry(self.fs, full_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[XfsDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for name, entry in self._resolve().entry.listdir().items():
            if name in (None, ".", ".."):
                continue

            # TODO: Separate INode and DirEntry in dissect.xfs
            yield XfsDirEntry(self.fs, self.path, name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.filetype == stat.S_IFDIR
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.filetype == stat.S_IFREG
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.filetype == stat.S_IFLNK

    def readlink(self) -> str:
        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        node = self.entry.inode

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                node.di_mode,
                self.entry.inum,
                id(self.fs),
                node.di_nlink,
                node.di_uid,
                node.di_gid,
                self.entry.size,
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = self.entry.atime_ns
        st_info.st_mtime_ns = self.entry.mtime_ns
        st_info.st_ctime_ns = self.entry.ctime_ns

        st_info.st_blksize = self.fs.xfs.block_size
        # Convert number of filesystem blocks to basic blocks
        # Reference: https://github.com/torvalds/linux/blob/e32cde8d2bd7d251a8f9b434143977ddf13dcec6/fs/xfs/xfs_iops.c#L602 # noqa: E501
        # Note that block size in XFS is always a multiple of 512, so the division below is safe
        st_info.st_blocks = self.entry.nblocks * (self.fs.xfs.block_size // 512)

        # XFS has a birth time, since inode version 3 (version 5 of filesystem)
        st_info.st_birthtime = self.entry.crtime.timestamp()
        st_info.st_birthtime_ns = self.entry.crtime_ns

        return st_info

    def attr(self) -> Any:
        return self._resolve().entry.xattr

    def lattr(self) -> Any:
        return self.entry.xattr



from __future__ import annotations

import stat
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.database.ese import ESE
from dissect.database.ese.c_ese import ulDAEMagic
from dissect.database.ese.ntds import NTDS, Object

from dissect.target.exceptions import FileNotFoundError, NotASymlinkError
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class NtdsFilesystem(Filesystem):
    """Filesystem implementation for NTDS.dit files. Because we can."""

    __type__ = "ntds"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        self.ntds = NTDS(fh)
        super().__init__(fh, *args, case_sensitive=False, **kwargs)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        buf = fh.read(8)
        if int.from_bytes(buf[4:8], "little") == ulDAEMagic:
            ese = ESE(fh)
            return {"datatable", "link_table"}.issubset(table.name for table in ese.tables())
        return False

    def get(self, path: str) -> NtdsFilesystemEntry:
        return NtdsFilesystemEntry(self, path, self._get_object(path))

    def _get_object(self, path: str, root: Object | None = None) -> Object:
        obj = root or self.ntds.root()

        for part in path.split("/"):
            if not part or part == ".":
                continue

            if part == "..":
                if obj.parent() is None:
                    continue
                obj = obj.parent()
            else:
                for child in obj.children():
                    if child.name == part:
                        obj = child
                        break
                else:
                    raise FileNotFoundError(f"Path not found: {path}")

        return obj


class NtdsDirEntry(DirEntry):
    entry: Object

    def get(self) -> NtdsFilesystemEntry:
        return NtdsFilesystemEntry(self.fs, self.path, self.entry)

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        return next(self.entry.children(), None) is not None

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        return True

    def is_symlink(self) -> bool:
        return False

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class NtdsFilesystemEntry(FilesystemEntry):
    fs: NtdsFilesystem
    entry: Object

    def get(self, path: str) -> NtdsFilesystemEntry:
        return NtdsFilesystemEntry(
            self.fs,
            fsutil.join(self.path, path, alt_separator=self.fs.alt_separator),
            self.fs._get_object(path, self.entry),
        )

    def open(self) -> BinaryIO:
        info = "\n".join(f"{key}: {value}" for key, value in self.entry.as_dict().items())
        return BytesIO(info.encode())

    def scandir(self) -> Iterator[NtdsDirEntry]:
        for child in self.entry.children():
            yield NtdsDirEntry(self.fs, self.path, child.name, child)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        return next(self.entry.children(), None) is not None

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return True

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError(self.path)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        mode = stat.S_IFDIR if self.is_dir() else stat.S_IFREG

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        return fsutil.stat_result(
            [
                mode | 0o777,
                self.entry.dnt,
                id(self.fs),
                1,
                0,
                0,
                0,
                0,
                self.entry.when_changed.timestamp() if self.entry.when_changed else 0,
                self.entry.when_created.timestamp() if self.entry.when_created else 0,
            ]
        )



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


@pytest.fixture
def mock_impacket(monkeypatch: pytest.MonkeyPatch) -> Iterator[MagicMock]:
    with monkeypatch.context() as m:
        if "dissect.target.loaders.smb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.loaders.smb")
        if "dissect.target.filesystems.smb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.filesystems.smb")

        mock_impacket = MagicMock()
        m.setitem(sys.modules, "impacket", mock_impacket)
        m.setitem(sys.modules, "impacket.dcerpc", mock_impacket.dcerpc)
        m.setitem(sys.modules, "impacket.dcerpc.v5", mock_impacket.dcerpc.v5)
        m.setitem(sys.modules, "impacket.dcerpc.v5.rpcrt", mock_impacket.dcerpc.v5.rpcrt)
        m.setitem(sys.modules, "impacket.nt_errors", mock_impacket.nt_errors)
        m.setitem(sys.modules, "impacket.smb", mock_impacket.smb)
        m.setitem(sys.modules, "impacket.smb3structs", mock_impacket.smb3structs)
        m.setitem(sys.modules, "impacket.smbconnection", mock_impacket.smbconnection)

        yield mock_impacket


@pytest.fixture
def mock_connection(mock_impacket: MagicMock) -> MagicMock:
    mock_connection = MagicMock()
    mock_impacket.smbconnection.SMBConnection.return_value = mock_connection

    return mock_connection


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_connection: MagicMock) -> None:
    """Test that we correctly use ``SmbLoader`` when opening a ``Target``."""
    from dissect.target.loaders.smb import SmbLoader

    path = "smb://user@host"
    with patch("dissect.target.target.Target.apply"):
        target = opener(path)
        assert isinstance(target._loader, SmbLoader)
        assert target.path == Path("host")


def test_loader(mock_impacket: MagicMock, mock_connection: MagicMock) -> None:
    from dissect.target.filesystems.smb import SmbFilesystem
    from dissect.target.loader import open as loader_open
    from dissect.target.loaders.smb import SmbLoader, SmbRegistry

    loader = loader_open("smb://user@host")
    assert isinstance(loader, SmbLoader)
    assert loader._conn is mock_connection

    mock_impacket.smbconnection.SMBConnection.assert_called_with(
        remoteName="host", remoteHost="host", myName=SmbLoader.MACHINE_NAME
    )
    mock_connection.login.assert_called_with(
        domain=".", user="user", password="", nthash=SmbLoader.EMPTY_NT, lmhash=SmbLoader.EMPTY_LM
    )

    mock_connection.listShares.return_value = [{"shi1_netname": "C$\x00"}]

    t = Target()
    loader.map(t)

    assert len(t.filesystems) == 1
    assert isinstance(t.fs.mounts["c:"], SmbFilesystem)
    assert t.fs.mounts["c:"].share_name == "C$"

    assert len(t._plugins) == 1
    assert isinstance(t._plugins[0], SmbRegistry)



from __future__ import annotations

from typing import BinaryIO

from dissect.util import cpio

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.fsutil import open_decompress


class CpioFilesystem(TarFilesystem):
    __type__ = "cpio"

    def __init__(self, fh: BinaryIO, base: str | None = None, *args, **kwargs):
        super().__init__(open_decompress(fileobj=fh), base, *args, tarinfo=cpio.CpioInfo, **kwargs)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a cpio file on a given file-like object."""
        return cpio.detect_header(open_decompress(fileobj=fh)) != cpio.FORMAT_CPIO_UNKNOWN



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.esxi._os import ESXiPlugin, _create_local_fs, _decrypt_crypto_util
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test__create_tar_fs_no_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", False),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_crypto_util") as decrypt_func,
    ):
        target_linux._name = "local"
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__create_tar_fs_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", True),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_envelope") as decrypt_func,
    ):
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__create_tar_fs_failed_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", True),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_envelope", side_effect=[NotImplementedError]),
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_crypto_util") as decrypt_func,
    ):
        target_linux._name = "local"
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__decrypt_crypto_not_local(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    target_linux._name = "not_local"
    with patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", False):
        assert _create_local_fs(target_linux, fs_unix.path(""), fs_unix.path("")) is None


def test__decrypt_crypto_local(fs_unix: VirtualFilesystem) -> None:
    with patch("dissect.target.plugins.os.unix.esxi._os.subprocess.run") as mocked_run:
        mocked_run.return_value.stdout = b"data"

        assert _decrypt_crypto_util(fs_unix.path("data")).read() == b"data"


def test_esxi_os_detection(target_bare: Target, fs_esxi: VirtualFilesystem) -> None:
    target_bare.filesystems.add(fs_esxi)
    target_bare.apply()

    assert ESXiPlugin.detect(target_bare)
    assert isinstance(target_bare._os, ESXiPlugin)
    assert target_bare.os == OperatingSystem.ESXI
    assert target_bare.hostname == "localhost"
    assert target_bare.version == "6.7.0"
    assert target_bare.ips == ["192.168.56.101"]


def test_esxi_os_creation_version_7(target_bare: Target) -> None:
    """Test handling of ``ESXiPlugin.create`` for ESXi 7 with separate partitions.

    Indirectly tests the ESXi configstore plugin.
    """

    fs1 = VirtualFilesystem()
    fs1.map_file_fh("boot.cfg", BytesIO(b"build=7.13.37-1.2.3.4\nmodules=example.v00 --- example.tgz\n"))
    fs1.map_file_fh("example.v00", BytesIO(b""))
    target_bare.filesystems.add(fs1)

    fs2 = VirtualFilesystem()
    fs2.map_file_fh("/etc/vmware/esx.conf", BytesIO(b'/resourceGroups/version = "7.13.37"\n'))
    fs2.map_file(
        "/var/lib/vmware/configstore/backup/current-store-1",
        absolute_path("_data/plugins/os/unix/esxi/current-store-1"),
    )
    target_bare.filesystems.add(fs2)

    target_bare.apply()

    assert ESXiPlugin.detect(target_bare)
    assert ESXiPlugin.create(target_bare, fs1)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

WgetHstsRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "apps/shell/wget/hsts",
    [
        ("datetime", "ts_created"),
        ("uri", "host"),
        ("varint", "port"),
        ("boolean", "explicit_port"),
        ("boolean", "include_subdomains"),
        ("datetime", "max_age"),
        ("path", "source"),
    ],
)


class WgetPlugin(Plugin):
    """Wget shell plugin."""

    __namespace__ = "wget"

    def __init__(self, target: Target):
        super().__init__(target)
        self.artifacts = list(self._find_artifacts())

    def _find_artifacts(self) -> Iterator[tuple[UserDetails, TargetPath]]:
        for user_details in self.target.user_details.all_with_home():
            if (hsts_file := user_details.home_path.joinpath(".wget-hsts")).exists():
                yield hsts_file, user_details

    def check_compatible(self) -> None:
        if not self.artifacts:
            raise UnsupportedPluginError("No .wget-hsts files found on target")

    @export(record=WgetHstsRecord)
    def hsts(self) -> Iterator[WgetHstsRecord]:
        """Yield domain entries found in wget HSTS files.

        When using the ``wget`` command-line utility, a file named ``.wget-hsts`` is created in the user's home
        directory by default. The ``.wget-hsts`` file records HTTP Strict Transport Security (HSTS) information for the
        websites visited by the user via ``wget``.

        References:
            - https://www.gnu.org/software/wget
            - https://gitlab.com/gnuwget/wget/-/blob/master/src/hsts.c
            - https://en.wikipedia.org/wiki/HTTP_Strict_Transport_Security

        Yields ``WgetHstsRecord`` records with the following fields:

        .. code-block:: text

            ts_created          (datetime):  When the host was first added to the HSTS file
            host                (uri):       The host that was accessed over TLS by wget
            explicit_port       (boolean):   If the TCP port for TLS should be checked
            include_subdomains  (boolean):   If subdomains are included in the HSTS check
            max_age             (datetime):  Time to live of the entry in the HSTS file
            source              (path):      Location of the .wget-hsts file
        """
        for hsts_file, user_details in self.artifacts:
            if not hsts_file.is_file():
                continue

            for line in hsts_file.open("rt").readlines():
                if not (line := line.strip()) or line.startswith("#"):
                    continue

                try:
                    host, port, subdomain_count, created, max_age = line.split("\t")
                    port = int(port)
                except ValueError as e:
                    self.target.log.warning("Unexpected wget hsts line in file: %s", hsts_file)
                    self.target.log.debug("", exc_info=e)
                    continue

                yield WgetHstsRecord(
                    ts_created=int(created),
                    host=host,
                    port=port if port != 0 else None,
                    explicit_port=port != 0,
                    include_subdomains=int(subdomain_count),
                    max_age=int(created) + int(max_age),
                    source=hsts_file,
                    _user=user_details.user,
                    _target=self.target,
                )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from dissect.thumbcache import Error, Thumbcache
from dissect.thumbcache.tools.extract_with_index import dump_entry_data_through_index

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath

GENERIC_THUMBCACHE_FIELDS = [
    ("string", "identifier"),
    ("string", "hash"),
    ("string", "extension"),
    ("varint", "data_size"),
    ("bytes", "header_checksum"),
    ("bytes", "data_checksum"),
    ("path", "path"),
]
IndexRecord = TargetRecordDescriptor(
    "windows/thumbcache/index",
    [
        ("string", "identifier"),
        ("boolean", "in_use"),
        ("varint", "flags"),
        ("datetime", "last_modified"),
        ("path", "path"),
    ],
)
ThumbcacheRecord = TargetRecordDescriptor("windows/thumbcache/thumbcache", GENERIC_THUMBCACHE_FIELDS)
IconcacheRecord = TargetRecordDescriptor("windows/thumbcache/iconcache", GENERIC_THUMBCACHE_FIELDS)


class ThumbcachePlugin(Plugin):
    """Windows thumbcache plugin."""

    __namespace__ = "thumbcache"

    def get_cache_paths(self) -> Iterator[TargetPath]:
        for user_details in self.target.user_details.all_with_home():
            cache_path = user_details.home_path / "appdata/local/microsoft/windows/explorer"
            if cache_path.exists():
                yield cache_path

    def check_compatible(self) -> None:
        for cache_path in self.get_cache_paths():
            if len(list(cache_path.glob("*_idx.db"))) > 0:
                return
        raise UnsupportedPluginError("There was no cache path for that plugin")

    def _create_entries(
        self, cache: Thumbcache, record_type: TargetRecordDescriptor
    ) -> Iterator[ThumbcacheRecord | IconcacheRecord | IndexRecord]:
        for path, entry in cache.entries():
            yield record_type(
                identifier=entry.identifier,
                hash=entry.hash,
                extension=entry.extension,
                header_checksum=entry.header_checksum,
                data_checksum=entry.data_checksum,
                path=path,
                data_size=len(entry.data),
            )
        for index_entry in cache.index_entries():
            yield IndexRecord(
                identifier=index_entry.identifier.hex(),
                in_use=index_entry.in_use(),
                flags=index_entry.flags,
                last_modified=index_entry.last_modified,
                path=cache.index_file,
            )

    def _parse_thumbcache(
        self,
        record_type: TargetRecordDescriptor,
        prefix: str,
        output_dir: Path | None,
    ) -> Iterator[ThumbcacheRecord | IconcacheRecord | IndexRecord]:
        for cache_path in self.get_cache_paths():
            try:
                if output_dir:
                    dump_entry_data_through_index(cache_path, output_dir, prefix)
                else:
                    cache = Thumbcache(cache_path, prefix=prefix)
                    yield from self._create_entries(cache, record_type)

            except Error as e:  # noqa: PERF203
                # A specific thumbcache exception occurred, log the error.
                self.target.log.error(e)  # noqa: TRY400
            except Exception as e:
                # A different exception occurred, log the exception.
                self.target.log.error("Error parsing thumbcache: %s", e)  # noqa: TRY400
                self.target.log.debug(e, exc_info=e)

    @arg("-o", "--output", dest="output_dir", type=Path, help="path to extract thumbcache thumbnails to")
    @export(record=[ThumbcacheRecord, IndexRecord])
    def thumbcache(self, output_dir: Path | None = None) -> Iterator[ThumbcacheRecord | IndexRecord]:
        """Yield thumbcache thumbnails."""
        yield from self._parse_thumbcache(ThumbcacheRecord, "thumbcache", output_dir)

    @arg("-o", "--output", dest="output_dir", type=Path, help="path to extract iconcache thumbnails to")
    @export(record=[IconcacheRecord, IndexRecord])
    def iconcache(self, output_dir: Path | None = None) -> Iterator[IconcacheRecord | IndexRecord]:
        """Yield iconcache thumbnails."""
        yield from self._parse_thumbcache(IconcacheRecord, "iconcache", output_dir)



from __future__ import annotations

import pytest
from dissect.evidence.ad1.ad1 import find_files

from dissect.target.filesystems.ad1 import AD1Filesystem
from dissect.target.helpers import keychain
from tests._utils import absolute_path


def test_ad1_encrypted() -> None:
    """Test if we can open an AD1 ADCRYPT encrypted image using the keychain."""

    path = absolute_path("_data/filesystems/ad1/encrypted-small.ad1")
    segments = find_files(path)

    with pytest.raises(ValueError, match=r"Failed to unlock ADCRYPT: no key\(s\) provided"):
        AD1Filesystem(segments)

    keychain.register_wildcard_value("invalid")

    with pytest.raises(ValueError, match=r"Failed to unlock ADCRYPT using provided key\(s\)"):
        AD1Filesystem(segments)

    keychain.register_wildcard_value("password")
    fs = AD1Filesystem(segments)

    assert not fs.ad1.is_locked()
    assert list(fs.get("C:/Users/user/Desktop/Data").iterdir()) == [
        "hello.txt",
        "philipp-pilz-QZ2EQuPpQJs-unsplash.jpg",
    ]



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualKey
from dissect.target.plugins.os.windows.credential.winlogon import WinlogonPlugin

if TYPE_CHECKING:
    from dissect.target.helpers.regutil import VirtualHive
    from dissect.target.target import Target


def test_windows_credential_winlogon(target_win: Target, hive_hklm: VirtualHive) -> None:
    """Test if we can find a Winlogon DefaultPassword entry in the registry."""

    winlogon_path = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"
    winlogon_key = VirtualKey(hive_hklm, winlogon_path)
    winlogon_key.timestamp = datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    winlogon_key.add_value("DefaultPassword", "password")
    hive_hklm.map_key(winlogon_path, winlogon_key)

    target_win.add_plugin(WinlogonPlugin)

    record = next(target_win.winlogon())

    assert record.ts_mtime == datetime(2025, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    assert record.password == "password"
